// CCFWrapper.cpp: implementation of the CCCFWrapper class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "CCFWrapper.h"
#include "ProgressDlg.h"
#include "CSFSection.h"
#include "SimpleTextFile.h"
#include "SNConfig.h"
#include "CSFAbstractTree.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

void WINAPI CDDispStrCallback(char *str);

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CCCFWrapper::CCCFWrapper(const CSNConfig *pConfig)
{
	m_pConfig = pConfig;

	m_hInstCCFDLL = NULL;
	m_szLastDecompileAndReadInput = NULL;
	m_szPotentialDecompileAndReadInput = NULL;
	m_uliLastDecompileAndReadInputWrite.QuadPart = 0;
	m_uliPotentialDecompileAndReadInputWrite.QuadPart = 0;
	m_pRootCSFSection = NULL;
	m_pCSFAbstractTree = NULL;
	m_pGetCCFVersionFunctionFunction = NULL;
	m_pSetDisplayStringCallbackFunction = NULL;
	m_pCCFRunDecompilerFunction0_93 = NULL;
	m_pCCFRunCompilerFunction0_93 = NULL;

	GetCCFDllPath();
	m_sCCFDLLVersion = "CCFDll not initialised";
	m_iCCFDLLVersionMajor = 0;
	m_iCCFDLLVersionMinor = 0;
}

CCCFWrapper::~CCCFWrapper()
{
	delete m_pRootCSFSection;
	m_pRootCSFSection = NULL;

	delete m_pCSFAbstractTree;
	m_pCSFAbstractTree = NULL;

	delete [] m_szLastDecompileAndReadInput;
	m_szLastDecompileAndReadInput = NULL;

	delete [] m_szPotentialDecompileAndReadInput;
	m_szPotentialDecompileAndReadInput = NULL;

	if (NULL != m_hInstCCFDLL)
	{
		FreeLibrary(m_hInstCCFDLL);
		m_hInstCCFDLL = NULL;
	}
}

const char *CCCFWrapper::GetCCFDllPath(void)
{
	if (m_sCCFDLLPath.empty())
	{
		// Check if there is a copy of the DLL in the executable directory.
		// This should override all other paths so that people can use a specific CCFDll
		// with Super-Nudelist, if required.
		const char *szExePath = m_pConfig->GetExeDirectoryPath(NULL);

		if (NULL != szExePath)
		{
			char *szPath = JoinPaths(szExePath, CCFDLL_FILENAME);

			if (NudelAccess(NULL, szPath, false))
			{
				m_sCCFDLLPath = szPath;
			}

			delete [] szPath;
		}
	}

	if (m_sCCFDLLPath.empty())
	{
		// Second, check the directory specified in the registry by the CCF Tools installer.

		string sToolsPath;

		if (LeetRegQueryStringValue(HKEY_LOCAL_MACHINE, CCFDLL_REG_PATH_ROOT, CCFDLL_REG_VALUE_CCFTOOLSPATH, &sToolsPath))
		{
			char *szPath = JoinPaths(sToolsPath.c_str(), CCFDLL_FILENAME);

			if (NudelAccess(NULL, szPath, false))
			{
				m_sCCFDLLPath = szPath;
			}

			delete [] szPath;
		}
	}

	if (m_sCCFDLLPath.empty())
	{
		// Finally, let LoadLibrary search the path.
		m_sCCFDLLPath = CCFDLL_FILENAME;
	}

	return(m_sCCFDLLPath.c_str());
}

bool CCCFWrapper::Init(CProgressDlg *pProgressDlg, CTypedError *pTError)
{
	bool bResult = true;

	if (NULL != m_hInstCCFDLL)
	{
		bResult = false;
		CTypedError::SafeSetSystemError(pTError, "Internal error: CCFWrapper::Init called twice");
	}
	else
	{
		CProgressDlg::SafeSetMessage(pProgressDlg, "Loading %s", GetCCFDllPath());

		m_hInstCCFDLL = LoadLibrary(GetCCFDllPath());

		if (NULL == m_hInstCCFDLL)
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Could not load %s", GetCCFDllPath());
		}
		else
		{
			// Note: Tables below fails to compile in VC5. Have to put pointers in properly typed
			// variables and then cast those variables to (FARPROC*). Works fine in VC6.

			NudelLPAFuncTable funcsVersion[] =
			{
				{ (FARPROC*)&m_pGetCCFVersionFunctionFunction,		GETCCFVERSION				},
				{ NULL,												NULL						}
			};

			if (!loadFunctionsFromTable(pTError, funcsVersion))
			{
				bResult = false;
			}
			else
			{
				m_sCCFDLLVersion = m_pGetCCFVersionFunctionFunction();

				if (2 != sscanf(m_sCCFDLLVersion.c_str(), "CCFDll %d.%d", &m_iCCFDLLVersionMajor, &m_iCCFDLLVersionMinor))
				{
					bResult = false;
					CTypedError::SafeSetMsgError(pTError, "Unrecognised CCFDll version string: %s", m_sCCFDLLVersion.c_str());
				}
				else if ( (CCFDLL_VERSION_MIN_MAJOR >  m_iCCFDLLVersionMajor)
				||		 ((CCFDLL_VERSION_MIN_MAJOR == m_iCCFDLLVersionMajor) && (CCFDLL_VERSION_MIN_MINOR > m_iCCFDLLVersionMinor)))
				{
					bResult = false;
					CTypedError::SafeSetMsgError(pTError, "CCFDll %d.%d required but got CCFDll %d.%d",
													CCFDLL_VERSION_MIN_MAJOR,
													CCFDLL_VERSION_MIN_MINOR,
													m_iCCFDLLVersionMajor,
													m_iCCFDLLVersionMinor);
				}
				else
				{
					NudelLPAFuncTable funcs0_93[] =
					{
						{ (FARPROC*)&m_pSetDisplayStringCallbackFunction,	SETDISPLAYSTRINGCALLBACK	},
						{ (FARPROC*)&m_pCCFRunDecompilerFunction0_93,		CCFRUNDECOMPILER			},
						{ (FARPROC*)&m_pCCFRunCompilerFunction0_93,			CCFRUNCOMPILER				},
						{ NULL,												NULL						}
					};

					if (!loadFunctionsFromTable(pTError, funcs0_93))
					{
						bResult = false;
					}
				}

				if (bResult)
				{
					m_pSetDisplayStringCallbackFunction(CDDispStrCallback);
				}
			}

			CProgressDlg::SafeSetMessage(pProgressDlg, "");

			if (!bResult)
			{
				FreeLibrary(m_hInstCCFDLL);
				m_hInstCCFDLL = NULL;
			}
		}
	}

	return(bResult);
}

bool CCCFWrapper::loadFunctionsFromTable(CTypedError *pTError, NudelLPAFuncTable *funcTable)
{
	bool bResult = true;

	for (int i=0; (NULL != funcTable[i].ppFunc) && (NULL != funcTable[i].funcName); i++)
	{
		(*(funcTable[i].ppFunc)) = (FARPROC)GetProcAddress(m_hInstCCFDLL, funcTable[i].funcName);

		if (NULL == (*(funcTable[i].ppFunc)))
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Could not find %s in %s", funcTable[i].funcName, GetCCFDllPath());
			break;
		}
	}

	return(bResult);
}

const char *CCCFWrapper::GetCCFDllVersionString(void)
{
	return(m_sCCFDLLVersion.c_str());
}

bool CCCFWrapper::CleanTempDir(CProgressDlg *pProgressDlg, CTypedError *pTError)
{
	bool bResult = false;

	CProgressDlg::SafeSetHeading(pProgressDlg, "Cleaning up temp directory");
	CProgressDlg::SafeSetMessage(pProgressDlg, "");

	// If we empty the temp dir we're wiping out the results of decompilation.
	// See needToDecompile and commitDecompilationNote.
	SetDecompileNeeded();

	const char *tempPath = m_pConfig->GetTempPath();

	if (NULL == tempPath)
	{
		CTypedError::SafeSetMsgError(pTError, "Configuration returns NULL temp path");
	}
	else
	{
		bool bWasError = CTypedError::SafeIsError(pTError);

		if (! NudelEmptyDirectory(pProgressDlg, tempPath, pTError) ) // Create or empty the temp directory.
		{
			if (!bWasError)
			{
				CTypedError::SafePrefixMsgError(pTError, "Failed to clean temp area");
			}
		}
		else
		{
			bResult = true;
		}
	}

	return(bResult);
}

// Returns success.
// Will only decompile and read if needed, i.e. if the input has changed since last time.
// pStrLstLog must NEVER be NULL.
// pStrLstLog is cleared at start of function and, if successful, cleared at the end.
bool CCCFWrapper::DecompileAndRead(CProgressDlg *pProgressDlg, CStringList *pStrLstLog, CTypedError *pTError)
{
	bool bResult = false;

	CProgressDlg::SafeSetHeading(pProgressDlg, "Preparing to decompile and read");
	CProgressDlg::SafeSetMessage(pProgressDlg, "");

	// Empty any old stuff in the log.
	pStrLstLog->RemoveAll();

	const char *inputCCFPath = m_pConfig->GetInputCCFPath();
	const char *tempCSFPath = m_pConfig->GetTempCSFPath();

	if (NULL == inputCCFPath)
	{
		CTypedError::SafeSetMsgError(pTError, "Configuration returns NULL input CCF path");
	}
	else if (NULL == tempCSFPath)
	{
		CTypedError::SafeSetMsgError(pTError, "Configuration returns NULL temp CSF path");
	}
	else
	{
		CProgressDlg::SafeSetMessage(pProgressDlg, "Evaluating need to decompile");

		if (!needToDecompileAndRead(inputCCFPath))
		{
			CProgressDlg::SafeSetMessage(pProgressDlg, "Decompilation unnecessary");
			bResult = true;
		}
		else
		{
			CProgressDlg::SafeDoImmediatePopup(pProgressDlg);

			delete m_pRootCSFSection;
			m_pRootCSFSection = NULL;

			delete m_pCSFAbstractTree;
			m_pCSFAbstractTree = NULL;

			if (CleanTempDir(pProgressDlg, pTError))
			{
				CProgressDlg::SafeSetHeading(pProgressDlg, "Decompiling");
				CProgressDlg::SafeSetMessage(pProgressDlg, "");

				if (actualDecompile(inputCCFPath, tempCSFPath, pProgressDlg, pStrLstLog, pTError))
				{
					// Clear the log since there was no error during decompilation.
					pStrLstLog->RemoveAll();

					CProgressDlg::SafeSetHeading(pProgressDlg, "Loading");
					CProgressDlg::SafeSetMessage(pProgressDlg, "Reading decompiled file from disk...");

					FILE *pFileInputCSF = fopen(tempCSFPath, "r");

					if (NULL == pFileInputCSF)
					{
						CTypedError::SafeSetMsgError(pTError, "Unable to open %s for reading.", tempCSFPath);
					}
					else
					{
						m_pRootCSFSection = new CCSFSection(NULL);
						m_pCSFAbstractTree = new CCSFAbstractTree();

						CSimpleTextFile *pTextFile = new CSimpleTextFile;

						if (! pTextFile->Read(pFileInputCSF) )
						{
							CTypedError::SafeSetMsgError(pTError, "Unable to read %s", tempCSFPath);
						}
						else
						{
							CProgressDlg::SafeSetMessage(pProgressDlg, "Parsing...");

							CSimpleTextFile::const_lineIterator lineIter = pTextFile->lineBegin();
							CSimpleTextFile::const_lineIterator lineIterEnd = pTextFile->lineEnd();

							if ((m_pRootCSFSection->ReadNextSection(lineIter, lineIterEnd, pTError))\
							&&	(m_pCSFAbstractTree->Build(pTError, m_pRootCSFSection))				)
							{
								// Commit candidate details.
								commitDecompileAndReadNote();

								bResult = true;
							}
						}

						delete pTextFile;
						pTextFile = NULL;

						if (!bResult)
						{
							delete m_pRootCSFSection;
							m_pRootCSFSection = NULL;

							delete m_pCSFAbstractTree;
							m_pCSFAbstractTree = NULL;
						}

						fclose(pFileInputCSF);
						pFileInputCSF = NULL;
					}
				}
			}
		}
	}

	return(bResult);
}

// Returns success.
// Should only be called after a successful decompilation.
// Rebuilds the Abstract Tree from the in-memory CSF data; never causes a new decompilation.
bool CCCFWrapper::RebuildAbstractTree(CTypedError *pTError)
{
	delete m_pCSFAbstractTree;
	m_pCSFAbstractTree = new CCSFAbstractTree();

	return(m_pCSFAbstractTree->Build(pTError, m_pRootCSFSection));
}

// Returns success.
// pStrLstLog must NEVER be NULL.
// pStrLstLog is cleared at start of function and, if successful, cleared at the end.
bool CCCFWrapper::SaveAndCompile(CProgressDlg *pProgressDlg, CStringList *pStrLstLog, CTypedError *pTError)
{
	bool bResult = false;

	CProgressDlg::SafeSetHeading(pProgressDlg, "Compiling");
	CProgressDlg::SafeSetMessage(pProgressDlg, "");

	// Empty any old stuff in the log.
	pStrLstLog->RemoveAll();

	const char *tempCSFPath = m_pConfig->GetTempCSFPath();
	const char *outputCCFPath = m_pConfig->GetOutputCCFPath();

	if (NULL == tempCSFPath)
	{
		CTypedError::SafeSetMsgError(pTError, "Configuration returns NULL temp CSF path");
	}
	else if (NULL == outputCCFPath)
	{
		CTypedError::SafeSetMsgError(pTError, "Configuration returns NULL output CCF path");
	}
	else
	{
		FILE *pFile = fopen(tempCSFPath, "w");

		if (NULL == pFile)
		{
			CTypedError::SafeSetMsgError(pTError, "Unable to write %s", tempCSFPath);
		}
		else
		{
			int nTotalCSFLines;

			if (! m_pRootCSFSection->Write(pFile, &nTotalCSFLines, 0) )
			{
				CTypedError::SafeSetMsgError(pTError, "Unable to write to %s", tempCSFPath);
				fclose(pFile);
			}
			else
			{
				fclose(pFile);

				if (actualCompile(tempCSFPath, outputCCFPath, nTotalCSFLines, pProgressDlg, pStrLstLog, pTError))
				{
					// Clear the log since there was no error during compilation.
					pStrLstLog->RemoveAll();

					bResult = true;
				}
			}
		}
	}

	return(bResult);
}

CCSFSection *CCCFWrapper::GetRootCSFSection(void)
{
	return(m_pRootCSFSection);
}

const CCSFSection *CCCFWrapper::GetRootCSFSection(void) const
{
	return(m_pRootCSFSection);
}

CCSFAbstractTree *CCCFWrapper::GetCSFAbstractTree(void)
{
	return(m_pCSFAbstractTree);
}

const CCSFAbstractTree *CCCFWrapper::GetCSFAbstractTree(void) const
{
	return(m_pCSFAbstractTree);
}

// Returns success.
bool CCCFWrapper::RemoveDanglingJumps(CTypedError *pTError, bool bRebuildAbstractTree)
{
	bool bResult = true;

	if (bRebuildAbstractTree)
	{
		bResult = RebuildAbstractTree(pTError);
	}

	if (bResult)
	{
		if (NULL == m_pRootCSFSection)
		{
			CTypedError::SafeSetMsgError(pTError, "RemoveDanglingJumps called but m_pRootCSFSection is NULL");
			bResult = false;
		}
		else if (NULL == m_pCSFAbstractTree)
		{
			CTypedError::SafeSetMsgError(pTError, "RemoveDanglingJumps called but m_pCSFAbstractTree is NULL");
			bResult = false;
		}
		else
		{
			bResult = removeDanglingJumps(pTError, m_pRootCSFSection);
		}
	}

	return(bResult);
}

bool CCCFWrapper::removeDanglingJumps(CTypedError *pTError, CCSFSection *pSection)
{
	bool bResult = true;

	if (NULL != pSection)
	{
		char *deviceNameBuffer;
		char *panelNameBuffer;

		bool bStillExists = true;

		list<const CCSFATPanel *> panelList;

		if (pSection->ScanForDevicePanelJumpEscaped(&deviceNameBuffer, &panelNameBuffer))
		{
			CProntoString psDeviceName;
			CProntoString psPanelName;
			psDeviceName.SetEscaped(deviceNameBuffer);
			psPanelName.SetEscaped(panelNameBuffer);

			m_pCSFAbstractTree->FindPanelsPronto(panelList, psDeviceName.GetPronto(false), psPanelName.GetPronto(false));

			bStillExists = false;

			// Make sure the match is exact as the CCFDll is case-sensitive.
			for (list<const CCSFATPanel *>::const_iterator liter = panelList.begin(); liter != panelList.end(); liter++)
			{
				if ((0 == (psPanelName.Compare((*liter)->GetName(), true)))
				&&	(0 == (psDeviceName.Compare((*liter)->GetParentDevice()->GetName(), true))))
				{
					bStillExists = true;
					break;
				}
			}

			if ((!bStillExists) && (NULL != pSection->GetParent()))
			{
				// Make it so!
				pSection->GetParent()->RemoveAndDeleteSubSectionByPointer(pSection);
			}

			delete [] deviceNameBuffer;
			delete [] panelNameBuffer;
		}

		if (bStillExists)
		{
			CCSFSection::subIterator iter = pSection->subBegin(); 

			while (iter != pSection->subEnd())
			{
				// We must increment the iterator before calling removeDanglingJumps since the call may delete
				// the element and thus invalidate the iterator.
				CCSFSection *pSubSection = (*(iter++));

				if (! removeDanglingJumps(pTError, pSubSection) )
				{
					bResult = false;
					break;
				}
			}
		}
	}

	return(bResult);
}

void CCCFWrapper::SetDecompileNeeded(void)
{
	// Delete any existing potential input.
	delete [] m_szLastDecompileAndReadInput;
	m_szLastDecompileAndReadInput = NULL;
}

// Returns true unless the input filename has just been decompiled and read and has not been changed since.
// Call commitDecompileAndReadNote() if we go on to successfully decompile and read the changed input.
bool CCCFWrapper::needToDecompileAndRead(const char *inputFilename)
{
	bool bResult = true;

	// Delete any existing potential input.
	delete [] m_szPotentialDecompileAndReadInput;
	m_szPotentialDecompileAndReadInput = NULL;

	m_szPotentialDecompileAndReadInput = new char[strlen(inputFilename) + 1];
	strcpy(m_szPotentialDecompileAndReadInput, inputFilename);

	HANDLE hFile;

	hFile = CreateFile(m_szPotentialDecompileAndReadInput, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
						FILE_ATTRIBUTE_NORMAL, NULL);

	// If we cannot open the file, assume we need to decompile again.
	// (Decompilation will likely fail, since we can't open the file. Let's report the problem there for simplicity.)
	if (INVALID_HANDLE_VALUE != hFile)
	{
		FILETIME ftWrite;

		// If we cannot get the write time, assume we need to decompile again.
		if (GetFileTime(hFile, NULL, NULL, &ftWrite))
		{
			m_uliPotentialDecompileAndReadInputWrite.LowPart = ftWrite.dwLowDateTime;
			m_uliPotentialDecompileAndReadInputWrite.HighPart = ftWrite.dwHighDateTime;

			// We have decompiled before...
			if (NULL != m_szLastDecompileAndReadInput)
			{
				// And it was the same filename...
				if (0 == strcmp(m_szLastDecompileAndReadInput, m_szPotentialDecompileAndReadInput))
				{
					// And the file hasn't been modified since...
					if (m_uliLastDecompileAndReadInputWrite.QuadPart == m_uliPotentialDecompileAndReadInputWrite.QuadPart)
					{
						// Then we don't need to decompile again.
						bResult = false;

						// Delete potential input since it's a waste of memory.
						delete [] m_szPotentialDecompileAndReadInput;
						m_szPotentialDecompileAndReadInput = NULL;
					}
				}
			}
		}

		CloseHandle(hFile);
	}

	return(bResult);
}

// Update internal variables to say that the file last given to needToDecompileAndRead has been
// successfully decompiled and read.
// Variables are used by needToDecompileAndRead() to see if the input has changed.
void CCCFWrapper::commitDecompileAndReadNote(void)
{
	if (NULL != m_szPotentialDecompileAndReadInput)
	{
		delete [] m_szLastDecompileAndReadInput;
		m_szLastDecompileAndReadInput = NULL;

		m_szLastDecompileAndReadInput = m_szPotentialDecompileAndReadInput;
		m_szPotentialDecompileAndReadInput = NULL;

		m_uliLastDecompileAndReadInputWrite.QuadPart = m_uliPotentialDecompileAndReadInputWrite.QuadPart;
	}
}

bool CCCFWrapper::actualDecompile(const char *inputCCFPath, const char *tempCSFPath,
							  CProgressDlg *pProgressDlg, CStringList *pStrLstLog, CTypedError *pTError)
{
	bool bResult = false;

	if (NULL == m_hInstCCFDLL)
	{
		CTypedError::SafeSetMsgError(pTError, "CCFDll not properly initialised");
	}
	else
	{
		// Since we cannot give the CCFDLL's progress function arguments we must use global
		// variables. Make sure only one thread at a time is using them.
		EnterCriticalSection(&g_CSGlobals);

		g_pProgressDlg = pProgressDlg;
		g_pStringList = pStrLstLog;
		g_bCompiling = false;
		g_nLastLineNumber = 0;

		int iFlags = CCF_OUTPUT_FORMAT_CSF;

		if ( ( CCFDLL_VERSION_095_MAJOR  < m_iCCFDLLVersionMajor)
		||	 ((CCFDLL_VERSION_095_MAJOR == m_iCCFDLLVersionMajor) && (CCFDLL_VERSION_095_MINOR <= m_iCCFDLLVersionMinor)))
		{
			iFlags |= CCF_OUTPUT_FORMAT_GENERATE_BUTTON_ID;
		}

		if (0 == m_pCCFRunDecompilerFunction0_93(const_cast<char *>(inputCCFPath), const_cast<char *>(tempCSFPath), iFlags))
		{
			if (pStrLstLog->IsEmpty())
			{
				CTypedError::SafeSetMsgError(pTError, "Decompilation failed; No log was received");
			}
			else
			{
				CTypedError::SafeSetMsgError(pTError, "DECOMPILATION FAILED, see log for details:");
			}
		}
		else
		{
			bResult = true;
		}

		g_pProgressDlg = NULL;
		g_pStringList = NULL;

		LeaveCriticalSection(&g_CSGlobals);
	}

	return(bResult);
}

bool CCCFWrapper::actualCompile(const char *tempCSFPath, const char *outputCCFPath, int nTotalCSFLines,
							  CProgressDlg *pProgressDlg, CStringList *pStrLstLog, CTypedError *pTError)
{
	bool bResult = false;

	if (NULL == m_hInstCCFDLL)
	{
		CTypedError::SafeSetMsgError(pTError, "CCFDll not properly initialised");
	}
	else
	{
		// Since we cannot give the CCFDLL's progress function arguments we must use global
		// variables. Make sure only one thread at a time is using them.
		EnterCriticalSection(&g_CSGlobals);

		g_pProgressDlg = pProgressDlg;
		g_pStringList = pStrLstLog;
		g_bCompiling = true;
		g_nLastLineNumber = 0;

		int iFlags = 0;

		if (m_pConfig->GetCompressImages())
		{
			iFlags |= CCF_COMPILER_FLAG_COMPRESS_ICON;
		}

		CProgressDlg::SafeSetNumbers(pProgressDlg, 0, nTotalCSFLines);

		if (0 == m_pCCFRunCompilerFunction0_93(const_cast<char *>(tempCSFPath), const_cast<char *>(outputCCFPath), iFlags))
		{
			if (pStrLstLog->IsEmpty())
			{
				CTypedError::SafeSetMsgError(pTError, "Compilation failed; No log was received");
			}
			else
			{
				CTypedError::SafeSetMsgError(pTError, "COMPILATION FAILED, see log for details:");
			}
		}
		else
		{
			bResult = true;
		}

		g_pProgressDlg = NULL;
		g_pStringList = NULL;

		LeaveCriticalSection(&g_CSGlobals);
	}

	return(bResult);
}

void WINAPI CDDispStrCallback(char *str)
{
	CString cstr = str;
	cstr.TrimLeft();
	cstr.TrimRight();

	int nLineNumber;

	if (g_bCompiling
	&&  ((1 == sscanf(LPCTSTR(cstr), "Analyzing CCF (line %d)",		&nLineNumber))
	||	 (1 == sscanf(LPCTSTR(cstr), "Analyzing Entry (line %d)",	&nLineNumber))
	||	 (1 == sscanf(LPCTSTR(cstr), "Analyzing Action (line %d)",	&nLineNumber))
	||	 (1 == sscanf(LPCTSTR(cstr), "Analyzing Panel (line %d)",	&nLineNumber))
	||	 (1 == sscanf(LPCTSTR(cstr), "Analyzing Frame (line %d)",	&nLineNumber))
	||	 (1 == sscanf(LPCTSTR(cstr), "Analyzing Button (line %d)",	&nLineNumber))))
	{
		if (nLineNumber > (g_nLastLineNumber + 1000))
		{
			CProgressDlg::SafeSetNumbers(g_pProgressDlg, nLineNumber, -1);
			CProgressDlg::SafeSetHeading(g_pProgressDlg, "Compiling line %d", nLineNumber);
			g_nLastLineNumber = nLineNumber;
		}
	}
	else if ((0 != stricmp(LPCTSTR(cstr), "Checking XML format"))
	&&		 (0 != stricmp(LPCTSTR(cstr), "Cannot recognize XML format"))
	&&		 (0 != stricmp(LPCTSTR(cstr.Right(33)), "Start tag expected, \"<\" not found")))
	{
		if (g_bCompiling)
		{
			CProgressDlg::SafeSetHeading(g_pProgressDlg, "Compiling", nLineNumber);
			g_nLastLineNumber = 0;
		}

		CProgressDlg::SafeSetMessage(g_pProgressDlg, "%s", LPCTSTR(cstr));

		if (NULL != g_pStringList)
		{
			g_pStringList->AddTail(cstr);
		}
	}
}
