// CSFPath.cpp: implementation of the CCSFPath class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "CSFPath.h"
#include "CSFSection.h"
#include "SNConfig.h"
#include "CSFAbstractTree.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//----------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------

CCSFPanelPath::CCSFPanelPath()
{
	m_separator = ", ";
}

CCSFPanelPath::~CCSFPanelPath()
{
}

void CCSFPanelPath::Reset(void)
{
	m_atPanels.clear();

	m_deviceName.Reset();
	m_panelName.Reset();

	m_description.Reset();
}

void CCSFPanelPath::Set(const CCSFPanelPath *pToCopy)
{
	if (this != pToCopy)
	{
		Reset();

		if (NULL != pToCopy)
		{
			m_atPanels.assign(pToCopy->m_atPanels.begin(), pToCopy->m_atPanels.end());

			m_deviceName.SetPronto(pToCopy->m_deviceName.GetPronto(false));
			m_panelName .SetPronto(pToCopy->m_panelName .GetPronto(false));
		}
	}
}

bool CCSFPanelPath::IsValid(CTypedError *pTError) const
{
	if (m_deviceName.IsUnassigned())
	{
		CTypedError::SafeSetMsgError(pTError, "panel path has no device name");
		return(false);
	}

	if (m_panelName.IsUnassigned())
	{
		CTypedError::SafeSetMsgError(pTError, "panel path has no panel name");
		return(false);
	}

	return(true);
}

const CProntoString *CCSFPanelPath::GetDeviceName(void) const
{
	return(&m_deviceName);
}

const CProntoString *CCSFPanelPath::GetPanelName(void) const
{
	return(&m_panelName);
}

void CCSFPanelPath::SetDeviceNameEscaped(const char *name)
{
	m_description.Reset();
	m_deviceName.SetEscaped(name);
}

void CCSFPanelPath::SetDeviceNamePronto(const char *name)
{
	m_description.Reset();
	m_deviceName.SetPronto(name);
}

void CCSFPanelPath::SetPanelNameEscaped(const char *name)
{
	m_description.Reset();
	m_panelName.SetEscaped(name);
}

void CCSFPanelPath::SetPanelNamePronto(const char *name)
{
	m_description.Reset();
	m_panelName.SetPronto(name);
}

const CProntoString *CCSFPanelPath::GetName(void) const
{
	if (m_description.IsUnassigned())
	{
		size_t descSize = 0;
		int elements = 0;

		if (! m_deviceName.IsUnassigned() )
		{
			elements++;
			descSize += strlen(m_deviceName.GetPronto(true));
		}

		if (! m_panelName.IsUnassigned() )
		{
			elements++;
			descSize += strlen(m_panelName.GetPronto(true));
		}

		if (0 == descSize)
		{
			m_description.SetPronto("<invalid CSF panel path>");
		}
		else
		{
			char *descPronto = new char[descSize + (elements * strlen(m_separator)) + 1];
			descPronto[0] = '\0';

			if (! m_deviceName.IsUnassigned() )
			{
				strcat(descPronto, m_deviceName.GetPronto(true));

				if (0 != --elements)
				{
					strcat(descPronto, m_separator);
				}
			}

			if (! m_panelName.IsUnassigned() )
			{
				strcat(descPronto, m_panelName.GetPronto(true));

				if (0 != --elements)
				{
					strcat(descPronto, m_separator);
				}
			}

			m_description.SetPronto(descPronto);

			delete [] descPronto;
		}
	}

	return(&m_description);
}

xmlNodePtr CCSFPanelPath::AddToXML(xmlNodePtr parentNode) const
{
	xmlNodePtr csfPathNode = xmlNewChild(parentNode, NULL, SNXML_NODE_CSFPATH, NULL);
	xmlSetProp(csfPathNode, SNXML_PROP_TYPE, SNXML_VALUE_PANEL);

	if (! m_deviceName.IsUnassigned() )
	{
		xmlSetProp(csfPathNode, SNXML_PROP_CSFDEVICENAME, (const xmlChar *)m_deviceName.GetEscaped(true));
	}

	if (! m_panelName.IsUnassigned() )
	{
		xmlSetProp(csfPathNode, SNXML_PROP_CSFPANELNAME, (const xmlChar *)m_panelName.GetEscaped(true));
	}

	return(csfPathNode);
}

bool CCSFPanelPath::BuildFromXML(CTypedError *pTError, CStringList *pSLErrorDetails, xmlNodePtr pNode)
{
	bool bResult = true;

	xmlChar *pXC;

	if (NULL == (pXC = xmlGetProp(pNode, SNXML_PROP_TYPE)))
	{
		CTypedError::SafeSetMsgError(pTError, "CSFPath has no type");
		bResult = false;
	}
	else
	{
		if (0 != strcmp((const char *)SNXML_VALUE_PANEL, (const char *)pXC))
		{
			CTypedError::SafeSetMsgError(pTError, "CSFPath type mismatch: expected \"%s\", got \"%s\"",
											SNXML_VALUE_PANEL, (const char *)pXC);
			bResult = false;
		}

		xmlFree(pXC);
	}

	if (bResult)
	{
		if (NULL != (pXC = xmlGetProp(pNode, SNXML_PROP_CSFDEVICENAME)))
		{
			this->SetDeviceNameEscaped((const char *)pXC);
			xmlFree(pXC);
		}

		if (NULL != (pXC = xmlGetProp(pNode, SNXML_PROP_CSFPANELNAME)))
		{
			this->SetPanelNameEscaped((const char *)pXC);
			xmlFree(pXC);
		}

		if (! this->IsValid(pTError) )
		{
			bResult = false;
		}
	}

	return(bResult);
}

bool CCSFPanelPath::ResolveCSFPath(CTypedError *pTError,
									bool bRequireUnique,
									const CCSFAbstractTree *pAT,
									const char *ownerType,
									const char *ownerName,
									const char *pathName)
{
	m_atPanels.clear();

	pAT->FindPanelsPronto(m_atPanels, m_deviceName.GetPronto(false),
										m_panelName.GetPronto(false));

	if (m_atPanels.begin() == m_atPanels.end())
	{
		CTypedError::SafeSetMsgError(pTError, "%s \"%s\" has non-existent %s",
												ownerType, ownerName, pathName);
		return(false);
	}
	
	if (bRequireUnique && ((++(m_atPanels.begin())) != m_atPanels.end()))
	{
		CTypedError::SafeSetMsgError(pTError, "%s \"%s\" has %s which is not unique",
												ownerType, ownerName, pathName);
		return(false);
	}

	bool bCaseSensitiveMatch = false;

	for (list<const CCSFATPanel *>::const_iterator liter = m_atPanels.begin(); liter != m_atPanels.end(); liter++)
	{
		if ((0 == ((*liter)->GetName()->Compare(&m_panelName, true)))
		&&	(0 == ((*liter)->GetParentDevice()->GetName()->Compare(&m_deviceName, true))))
		{
			bCaseSensitiveMatch = true;
			break;
		}
	}

	if (!bCaseSensitiveMatch)
	{
		CTypedError::SafeSetMsgError(pTError, "%s \"%s\" has %s which does not exist by that exact, case-sensitive name",
												ownerType, ownerName, pathName);
		return(false);
	}

	return(true);
}

// Only call after a successful ResolveCSFPath call.
const list<const CCSFATPanel *> *CCSFPanelPath::GetATPanels(void) const
{
	return(&m_atPanels);
}

//----------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------

CCSFButtonPath::CCSFButtonPath()
{
	m_separator = ", ";
}

CCSFButtonPath::~CCSFButtonPath()
{
}

void CCSFButtonPath::Reset(void)
{
	m_atButtons.clear();

	m_deviceName.Reset();
	m_panelName.Reset();
	m_buttonName.Reset();

	m_description.Reset();
}

void CCSFButtonPath::Set(const CCSFButtonPath *pToCopy)
{
	if (this != pToCopy)
	{
		Reset();

		if (NULL != pToCopy)
		{
			m_atButtons.assign(pToCopy->m_atButtons.begin(), pToCopy->m_atButtons.end());

			m_deviceName.SetPronto(pToCopy->m_deviceName.GetPronto(false));
			m_panelName .SetPronto(pToCopy->m_panelName .GetPronto(false));
			m_buttonName.SetPronto(pToCopy->m_buttonName.GetPronto(false));
		}
	}
}

const CProntoString *CCSFButtonPath::GetDeviceName(void) const
{
	return(&m_deviceName);
}

const CProntoString *CCSFButtonPath::GetPanelName(void) const
{
	return(&m_panelName);
}

const CProntoString *CCSFButtonPath::GetButtonName(void) const
{
	return(&m_buttonName);
}

void CCSFButtonPath::SetDeviceNameEscaped(const char *name)
{
	m_description.Reset();
	m_deviceName.SetEscaped(name);
}

void CCSFButtonPath::SetDeviceNamePronto(const char *name)
{
	m_description.Reset();
	m_deviceName.SetPronto(name);
}

void CCSFButtonPath::SetPanelNameEscaped(const char *name)
{
	m_description.Reset();
	m_panelName.SetEscaped(name);
}

void CCSFButtonPath::SetPanelNamePronto(const char *name)
{
	m_description.Reset();
	m_panelName.SetPronto(name);
}

void CCSFButtonPath::SetButtonNameEscaped(const char *name)
{
	m_description.Reset();
	m_buttonName.SetEscaped(name);
}

void CCSFButtonPath::SetButtonNamePronto(const char *name)
{
	m_description.Reset();
	m_buttonName.SetPronto(name);
}

const CProntoString *CCSFButtonPath::GetName(void) const
{
	if (m_description.IsUnassigned())
	{
		size_t descSize = 0;
		int elements = 0;

		if (! m_deviceName.IsUnassigned() )
		{
			elements++;
			descSize += strlen(m_deviceName.GetPronto(true));
		}

		if (! m_panelName.IsUnassigned() )
		{
			elements++;
			descSize += strlen(m_panelName.GetPronto(true));
		}

		if (! m_buttonName.IsUnassigned() )
		{
			elements++;
			descSize += strlen(m_buttonName.GetPronto(true));
		}

		if (0 == descSize)
		{
			m_description.SetPronto("<invalid CSF button path>");
		}
		else
		{
			char *descPronto = new char[descSize + (elements * strlen(m_separator)) + 1];
			descPronto[0] = '\0';

			if (! m_deviceName.IsUnassigned() )
			{
				strcat(descPronto, m_deviceName.GetPronto(true));

				if (0 != --elements)
				{
					strcat(descPronto, m_separator);
				}
			}

			if (! m_panelName.IsUnassigned() )
			{
				strcat(descPronto, m_panelName.GetPronto(true));

				if (0 != --elements)
				{
					strcat(descPronto, m_separator);
				}
			}

			if (! m_buttonName.IsUnassigned() )
			{
				strcat(descPronto, m_buttonName.GetPronto(true));

				if (0 != --elements)
				{
					strcat(descPronto, m_separator);
				}
			}

			m_description.SetPronto(descPronto);

			delete [] descPronto;
		}
	}

	return(&m_description);
}

bool CCSFButtonPath::IsValid(CTypedError *pTError) const
{
	if (m_deviceName.IsUnassigned())
	{
		CTypedError::SafeSetMsgError(pTError, "button path has no device name");
		return(false);
	}

	if (m_buttonName.IsUnassigned())
	{
		CTypedError::SafeSetMsgError(pTError, "button path has no button name");
		return(false);
	}

	if ((m_panelName.IsUnassigned()) && (!CCSFSection::IsHardButtonName(m_buttonName.GetPronto(false))))
	{
		CTypedError::SafeSetMsgError(pTError, "button path has no panel name and isn't a hard button");
		return(false);
	}

	return(true);
}

// Returns < 0 if this <  pButtonPath
// Returns   0 if this == pButtonPath
// Returns > 0 if this >  pButtonPath
int CCSFButtonPath::CompareForSort(const CCSFButtonPath *pButtonPath) const
{
	if (NULL == pButtonPath)
	{
		return(1);
	}
	else
	{
		return(GetName()->Compare(pButtonPath->GetName(), false));
	}
}

const char *CCSFButtonPath::getTypeName(void) const
{
	return("button path");
}

xmlNodePtr CCSFButtonPath::AddToXML(xmlNodePtr parentNode) const
{
	xmlNodePtr csfPathNode = xmlNewChild(parentNode, NULL, SNXML_NODE_CSFPATH, NULL);
	xmlSetProp(csfPathNode, SNXML_PROP_TYPE, SNXML_VALUE_BUTTON);

	if (! m_deviceName.IsUnassigned() )
	{
		xmlSetProp(csfPathNode, SNXML_PROP_CSFDEVICENAME, (const xmlChar *)m_deviceName.GetEscaped(true));
	}

	if (! m_panelName.IsUnassigned() )
	{
		xmlSetProp(csfPathNode, SNXML_PROP_CSFPANELNAME, (const xmlChar *)m_panelName.GetEscaped(true));
	}

	if (! m_buttonName.IsUnassigned() )
	{
		xmlSetProp(csfPathNode, SNXML_PROP_CSFBUTTONNAME, (const xmlChar *)m_buttonName.GetEscaped(true));
	}

	return(csfPathNode);
}

bool CCSFButtonPath::BuildFromXML(CTypedError *pTError, CStringList *pSLErrorDetails, xmlNodePtr pNode)
{
	bool bResult = true;

	xmlChar *pXC;

	if (NULL == (pXC = xmlGetProp(pNode, SNXML_PROP_TYPE)))
	{
		CTypedError::SafeSetMsgError(pTError, "CSFPath has no type");
		bResult = false;
	}
	else
	{
		if (0 != strcmp((const char *)SNXML_VALUE_BUTTON, (const char *)pXC))
		{
			CTypedError::SafeSetMsgError(pTError, "CSFPath type mismatch: expected \"%s\", got \"%s\"",
											SNXML_VALUE_BUTTON, (const char *)pXC);
			bResult = false;
		}

		xmlFree(pXC);
	}

	if (bResult)
	{
		if (NULL != (pXC = xmlGetProp(pNode, SNXML_PROP_CSFDEVICENAME)))
		{
			this->SetDeviceNameEscaped((const char *)pXC);
			xmlFree(pXC);
		}

		if (NULL != (pXC = xmlGetProp(pNode, SNXML_PROP_CSFPANELNAME)))
		{
			this->SetPanelNameEscaped((const char *)pXC);
			xmlFree(pXC);
		}

		if (NULL != (pXC = xmlGetProp(pNode, SNXML_PROP_CSFBUTTONNAME)))
		{
			this->SetButtonNameEscaped((const char *)pXC);
			xmlFree(pXC);
		}

		if (! this->IsValid(pTError) )
		{
			bResult = false;
		}
	}

	return(bResult);
}

bool CCSFButtonPath::ResolveCSFPath(CTypedError *pTError,
								 bool bRequireUnique,
								 const CCSFAbstractTree *pAT,
								 const char *ownerType,
								 const char *ownerName,
								 const char *pathName)
{
	m_atButtons.clear();

	pAT->FindButtonsPronto(m_atButtons, m_deviceName.GetPronto(false),
										m_panelName.GetPronto(false),
										m_buttonName.GetPronto(false));

	if (m_atButtons.begin() == m_atButtons.end())
	{
		CTypedError::SafeSetMsgError(pTError, "%s \"%s\" has non-existent %s",
												ownerType, ownerName, pathName);
		return(false);
	}

	if (bRequireUnique && ((++(m_atButtons.begin())) != m_atButtons.end()))
	{
		CTypedError::SafeSetMsgError(pTError, "%s \"%s\" has %s which is not unique",
												ownerType, ownerName, pathName);
		return(false);
	}

	bool bCaseSensitiveMatch = false;

	for (list<const CCSFATButton *>::const_iterator liter = m_atButtons.begin(); liter != m_atButtons.end(); liter++)
	{
		if (0 == ((*liter)->GetName()->Compare(&m_buttonName, true)))
		{
			const CCSFATPanel *pParentPanelIfSoft = (*liter)->GetParentPanelIfSoft();

			if (NULL != pParentPanelIfSoft)
			{
				if ((0 == (pParentPanelIfSoft->GetName()->Compare(&m_panelName, true)))
				&&	(0 == (pParentPanelIfSoft->GetParentDevice()->GetName()->Compare(&m_deviceName, true))))
				{
					bCaseSensitiveMatch = true;
					break;
				}

			}
			else
			{
				const CCSFATDevice *pParentDeviceIfHard = (*liter)->GetParentDeviceIfHard();

				if (NULL != pParentDeviceIfHard)
				{
					if (0 == (pParentDeviceIfHard->GetName()->Compare(&m_deviceName, true)))
					{
						bCaseSensitiveMatch = true;
						break;
					}
				}
			}
		}
	}

	if (!bCaseSensitiveMatch)
	{
		CTypedError::SafeSetMsgError(pTError, "%s \"%s\" has %s which does not exist by that exact, case-sensitive name",
												ownerType, ownerName, pathName);
		return(false);
	}

	return(true);
}

// Only call after a successful ResolveCSFPath call.
const list<const CCSFATButton *> *CCSFButtonPath::GetATButtons(void) const
{
	return(&m_atButtons);
}

//----------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------
