
#include "stdafx.h"
#include "CSFSection.h"
#include "UnsignedBaseTenNumber.h"

/////////////////////////////////////////////////////////////////////////

CCSFSection::CCSFSection(CCSFSection *pSectionParent)
{
	m_pSectionParent = pSectionParent;

	if (NULL != m_pSectionParent)
	{
		m_pSectionParent->m_bHasSubBody = true;
		m_pSectionParent->m_subSections.push_back(this);
		m_pHighestButtonNumber = m_pSectionParent->m_pHighestButtonNumber;
	}
	else
	{
		m_pHighestButtonNumber = new (unsigned int);
		(*m_pHighestButtonNumber) = 0;
	}

	m_title = NULL;
	m_bHasSubBody = false;
}

CCSFSection::~CCSFSection(void)
{
	SetTitleEscaped(NULL);

	while (! m_subSections.empty() )
	{
		delete m_subSections.back();
		m_subSections.pop_back();
	}

	if (NULL == m_pSectionParent)
	{
		// This is the root and thus it allocated it and must free it.
		delete m_pHighestButtonNumber;
		m_pHighestButtonNumber = NULL;
	}
}

CCSFSection *CCSFSection::GetParent(void)
{
	return(m_pSectionParent);
}

const CCSFSection *CCSFSection::GetParent(void) const
{
	return(m_pSectionParent);
}

const char *CCSFSection::GetTitleEscaped(void)
{
	return(m_title);
}

void CCSFSection::SetTitleEscaped(const char *title)
{
	delete [] m_title;
	m_title = NULL;

	if (NULL != title)
	{
		m_title = new char[strlen(title) + 1];
		strcpy(m_title, title);
	}
}

void CCSFSection::SetTitleEscapedFormat(const char *format, ...)
{
	if (NULL == format)
	{
		SetTitleEscaped(NULL);
	}
	else
	{
		char buffer[BUFFERSIZE];

		va_list args;
		va_start(args, format);
		vStringFormat(buffer, sizeof(buffer), format, args);
		va_end(args);

		SetTitleEscaped(buffer);
	}
}

CCSFSection::subIterator CCSFSection::subBegin(void)
{
	return(m_subSections.begin());
}

CCSFSection::subIterator CCSFSection::subEnd(void)
{
	return(m_subSections.end());
}

CCSFSection::const_subIterator CCSFSection::subBegin(void) const
{
	return(m_subSections.begin());
}

CCSFSection::const_subIterator CCSFSection::subEnd(void) const
{
	return(m_subSections.end());
}

bool CCSFSection::ReadNextSection(CSimpleTextFile::const_lineIterator &lineIter,
								  const CSimpleTextFile::const_lineIterator &lineIterEnd,
								  CTypedError *pTError)
{
	bool bResult = true;

	const char *lineBuffer;


	// Skip empty lines. (Whitespace has already been trimmed.)
	while(true)
	{
		if (lineIterEnd == lineIter)
		{
			CTypedError::SafeSetMsgError(pTError, "End of input CSF file reached unexpectedly\n");
			bResult = false;
			break;
		}
		else
		{
			lineBuffer = (*lineIter)->GetLineBuffer();

			if (0 != strlen(lineBuffer))
			{
				break;
			}
		}
	}

	if (bResult)
	{
		SetTitleEscaped(lineBuffer);

		lineIter++;

		if ((lineIterEnd != lineIter) && ('{' == (*lineIter)->GetLineBuffer()[0]))
		{
			lineIter++;

			m_bHasSubBody = true; // So we output the braces even if the subbody is empty.

			CCSFSection *pNewSubSection;

			while(true)
			{
				if (lineIterEnd == lineIter)
				{
					CTypedError::SafeSetMsgError(pTError, "End of input CSF file reached unexpectedly\n");
					bResult = false;
					break;
				}
				else
				{
					lineBuffer = (*lineIter)->GetLineBuffer();

					if (0 == strlen(lineBuffer)) // avoids missing the closing "}" if a blank line is right before it.
					{
						lineIter++;
					}
					else
					{
						if ('{' == lineBuffer[0])
						{
							CTypedError::SafeSetMsgError(pTError, "Unexpected '{' in input CSF file\n");
							bResult = false;
							break;
						}
						else if ('}' == lineBuffer[0])
						{
							// Finished reading the subsection.
							lineIter++;
							break;
						}
						else
						{
							pNewSubSection = AddSubSection();

							if (! pNewSubSection->ReadNextSection(lineIter, lineIterEnd, pTError) )
							{
								bResult = false;
								break; // LD 06/Sep/2000
							}
						}
					}
				}
			}
		}
	}

	return(bResult);
}

// Set *pLineCount to the number of lines written.
bool CCSFSection::Write(FILE *pfOutputCSF, int *pLineCount, int indentLevel) const
{
	if (NULL != pLineCount)
	{
		*pLineCount = 0;
	}

	if ((NULL != m_title) && ('\0' != m_title[0]))
	{
		indentLine(pfOutputCSF, indentLevel);

		fprintf(pfOutputCSF, "%s\n", (const char *)m_title);

		if (NULL != pLineCount) { (*pLineCount)++; }

		if (m_bHasSubBody)
		{
			indentLine(pfOutputCSF, indentLevel);
			fprintf(pfOutputCSF, "{\n");

			if (NULL != pLineCount) { (*pLineCount)++; }

			for (const_subIterator iter = subBegin(); subEnd() != iter; iter++)
			{
				int nSubLineCount;

				(*iter)->Write(pfOutputCSF, &nSubLineCount, indentLevel + 1);

				if (NULL != pLineCount) { (*pLineCount) += nSubLineCount; }

			}

			indentLine(pfOutputCSF, indentLevel);
			fprintf(pfOutputCSF, "}\n");

			if (NULL != pLineCount) { (*pLineCount)++; }
		}
	}

	return(true);
}

void CCSFSection::indentLine(FILE *pfOutputCSF, int indentLevel) const
{
	for (unsigned int i=0; i < indentLevel; i++)
	{
		fprintf(pfOutputCSF, "\t");
//		fprintf(pfOutputCSF, "  ");
	}
}

// Do not delete the subsection once it has been added, the class will on destruction.
// Do not add the same subsection pointer to more than one CCSFSection.
CCSFSection *CCSFSection::AddSubSection(void)
{
	CCSFSection *result = new CCSFSection(this);

	return(result);
}

CCSFSection *CCSFSection::AddEscapedTitleOnlySubSection(const char *subTitle)
{
	CCSFSection *result = AddSubSection();

	result->SetTitleEscaped(subTitle);

	return(result);
}

CCSFSection *CCSFSection::AddEscapedTitleOnlySubSectionFormat(const char *format, ...)
{
	CCSFSection *result = AddSubSection();

	if (NULL != format)
	{
		char buffer[BUFFERSIZE];

		va_list args;
		va_start(args, format);
		vStringFormat(buffer, sizeof(buffer), format, args);
		va_end(args);

		result->SetTitleEscaped(buffer);
	}

	return(result);
}

// label can be NULL.
// The subsection is returned so it can have further details added when required.
CCSFSection *CCSFSection::AddEscapedFrameSubSection(const char *label,
													unsigned int xpos, unsigned int ypos,
													unsigned int width, unsigned int height,
													CCSFSection::FontType fontType,
													CCSFSection::Colour labelTextColour,
													const char *iconPath)
{
	CCSFSection *result = AddSubSection();

	// If there's no label, set to FONT_0, else the Pronto will not draw the frame (ffs!).
	if ((NULL == label) || ('\0' == label[0]))
	{
		fontType = FONT_0;
	}

	result->SetTitleEscapedFormat("Frame(\"%s\",%u,%u,%u,%u)", NULL == label ? "" : label, xpos, ypos, width, height);
	result->AddEscapedTitleOnlySubSectionFormat("TextColor(%s)", getColourString(labelTextColour));
	result->AddEscapedTitleOnlySubSectionFormat("Font(%u);", (unsigned int)(fontType));

	if ((NULL != iconPath) && (0 < strlen(iconPath)))
	{
		result->AddEscapedTitleOnlySubSectionFormat("Icon(\"%s\");", iconPath);
	}

	return(result);
}

// Call this on the root before adding any buttons; it is NOT calculated automatically.
void CCSFSection::CalcHighestButtonNumber(void)
{
	if (NULL == m_pSectionParent)
	{
		// This is the root so reset the number.
		(*m_pHighestButtonNumber) = 0;
	}

	unsigned int buttonNumber;

	if (ScanForSoftButtonEscaped(NULL, &buttonNumber))
	{
		if ((*m_pHighestButtonNumber) < buttonNumber)
		{
			(*m_pHighestButtonNumber) = buttonNumber;
		}
	}

	for (subIterator iter = subBegin(); iter != subEnd(); iter++)
	{
		(*iter)->CalcHighestButtonNumber();
	}
}

// See SetHighestButtonNumber
// label can be NULL.
// The subsection is returned so it can have further details added when required.
CCSFSection *CCSFSection::AddEscapedButtonSubSection(const char *label,
														unsigned int xpos, unsigned int ypos,
														unsigned int width, unsigned int height,
														CCSFSection::FontType fontType,
														CCSFSection::Colour labelTextColour,
														const char *iconPath,
														CCSFSection **ppActionSection)
{
	CCSFSection *result = AddSubSection();

	unsigned int buttonNumber = ++(*m_pHighestButtonNumber);

	result->SetTitleEscapedFormat("Button(\"%s\",%u,%u,%u,%u,%u) /* Button %u */",
															NULL == label ? "" : label, xpos, ypos, width, height,
															buttonNumber, buttonNumber);

	result->AddEscapedTitleOnlySubSectionFormat("ActiveUnselectedColor(LightGray,%s);", getColourString(labelTextColour));
	result->AddEscapedTitleOnlySubSectionFormat("ActiveSelectedColor(LightGray,%s);", getColourString(labelTextColour));
	result->AddEscapedTitleOnlySubSectionFormat("Font(%u);", (unsigned int)(fontType));
	result->AddEscapedTitleOnlySubSectionFormat("ActiveIcons(\"%s\");", iconPath);

	CCSFSection *pActionSection = result->AddSubSection();
	pActionSection->SetTitleEscaped("Action()");

	if (NULL != ppActionSection)
	{
		(*ppActionSection) = pActionSection;
	}

	return(result);
}

// Returns NULL if there is no matching subsection.
CCSFSection *CCSFSection::FindSubSectionByNameEscaped(const char *name) const
{
	CCSFSection *result = NULL;

	const char *candidateName;

	if (m_bHasSubBody)
	{
		for (const_subIterator iter = subBegin(); subEnd() != iter; iter++)
		{
			candidateName = (*iter)->GetTitleEscaped();

			if (((NULL == name) && (NULL == candidateName))
			||  ((NULL != name) && (NULL != candidateName) && (0 == _stricmp(name, candidateName))))
			{
				result = (*iter);
				break;
			}
		}
	}

	return(result);
}

CCSFSection *CCSFSection::FindSubSectionByNamePrefixEscaped(const char *namePrefix) const
{
	CCSFSection *result = NULL;

	const char *candidateName;

	if (m_bHasSubBody)
	{
		size_t namePrefixLen = strlen(namePrefix);

		for (const_subIterator iter = subBegin(); subEnd() != iter; iter++)
		{
			candidateName = (*iter)->GetTitleEscaped();

			if (((NULL == namePrefix) && (NULL == candidateName))
			||  ((NULL != namePrefix) && (NULL != candidateName) && (0 == _strnicmp(namePrefix, candidateName, namePrefixLen))))
			{
				result = (*iter);
				break;
			}
		}
	}

	return(result);
}

void CCSFSection::RemoveAndDeleteSubSectionByPointer(CCSFSection *pSectionToFind)
{
	if (m_bHasSubBody)
	{
		m_subSections.remove(pSectionToFind); // Remove by value.
		delete pSectionToFind;

		// Do *not* update m_bHasSubBody. If m_subSections is now empty we have an empty subbody.
	}
}

const char *CCSFSection::getColourString(CCSFSection::Colour kula) const
{
	const char *result;

	switch(kula)
	{
	default:
	case(CCSFSection::COLOUR_BLACK):
		result = "Black";
		break;
	case(CCSFSection::COLOUR_DARKGRAY):
		result = "DarkGray";
		break;
	case(CCSFSection::COLOUR_LIGHTGRAY):
		result = "LightGray";
		break;
	case(CCSFSection::COLOUR_WHITE):
		result = "White";
		break;
	}

	return(result);
}

// Returns true if a button was parsed and the output arguments are valid.
// If returns true and pNameBuffer not NULL on input, *pNameBuffer is the name and should be delete[]'d.
// If returns true and pNumber not NULL on input, *pNumber is the number.
bool CCSFSection::ScanForSoftButtonEscaped(char **pNameBuffer, unsigned int *pNumber) const
{
	bool bResult = false;

	// Required for failure cleanup code.
	if (NULL != pNameBuffer)
	{
		(*pNameBuffer) = NULL;
	}

	const char *sectionTitle = m_title;

	if (NULL != sectionTitle)
	{
		const char *buttonPrefix = "Button(";
		size_t buttonPrefixLength = strlen(buttonPrefix);

		if (0 == _strnicmp(sectionTitle, buttonPrefix, buttonPrefixLength))
		{
			sectionTitle += buttonPrefixLength;

			size_t tokenBufferSize = strlen(sectionTitle) + 1;

			const char *bufPtr = sectionTitle;
			char *tokenBuffer = new char[tokenBufferSize];

			// Button name is the first argument
			// Button number is the sixth argument
			if (Tokenize(&bufPtr, ",", ")", tokenBuffer, tokenBufferSize, false, true, false))
			{
				if (NULL != pNameBuffer)
				{
					(*pNameBuffer) = new char[strlen(tokenBuffer) + 1];
					strcpy(*pNameBuffer, tokenBuffer);
				}

				if (Tokenize(&bufPtr, ",", ")", tokenBuffer, tokenBufferSize, false, true, false)	\
				&&  Tokenize(&bufPtr, ",", ")", tokenBuffer, tokenBufferSize, false, true, false)	\
				&&  Tokenize(&bufPtr, ",", ")", tokenBuffer, tokenBufferSize, false, true, false)	\
				&&  Tokenize(&bufPtr, ",", ")", tokenBuffer, tokenBufferSize, false, true, false)	\
				&&  Tokenize(&bufPtr, ",", ")", tokenBuffer, tokenBufferSize, false, true, false)	)
				{
					unsigned int buttonNumber = 0;
					
					if ((1 == sscanf(tokenBuffer, "%u", &buttonNumber)) && (0 != buttonNumber))
					{
						if (NULL != pNumber)
						{
							(*pNumber) = buttonNumber;
						}

						bResult = true; // Looks like a button to me.
					}
				}
			}

			delete [] tokenBuffer;
		}
	}

	// Failure cleanup code.
	if (!bResult)
	{
		// If we're returning false, free partial results.

		if ((NULL != pNameBuffer) && (NULL != (*pNameBuffer)))
		{
			delete [] (*pNameBuffer);
			(*pNameBuffer) = NULL;
		}
	}

	return(bResult);
}

// Returns true if a hard button was parsed and the output arguments are valid.
// If returns true and pNameBuffer not NULL on input, *pNameBuffer is the name and should be delete[]'d.
bool CCSFSection::ScanForHardButtonEscaped(char **pNameBuffer) const
{
	bool bResult = false;

	// Required for failure cleanup code.
	if (NULL != pNameBuffer)
	{
		(*pNameBuffer) = NULL;
	}

	const char *sectionTitle = m_title;

	if (NULL != sectionTitle)
	{
		const char *pOpenBracket = strchr(sectionTitle, '(');

		if (NULL != pOpenBracket)
		{
			size_t nameLen = (pOpenBracket - sectionTitle) + 1;
			char *name = new char[nameLen];
			stringCopy(name, sectionTitle, nameLen);

			if (IsHardButtonName(name))
			{
				bResult = true;

				if (NULL != pNameBuffer)
				{
					(*pNameBuffer) = name;
					name = NULL;
				}
			}

			delete [] name;
			name = NULL;
		}
	}

	// Failure cleanup code.
	if (!bResult)
	{
		// If we're returning false, free partial results.

		if ((NULL != pNameBuffer) && (NULL != (*pNameBuffer)))
		{
			delete [] (*pNameBuffer);
			(*pNameBuffer) = NULL;
		}
	}

	return(bResult);
}

// Returns true if a device was parsed and the output arguments are valid.
// If returns true and pNameBuffer not NULL on input, *pNameBuffer is the name and should be delete[]'d.
bool CCSFSection::ScanForDeviceEscaped(char **pNameBuffer) const
{
	bool bResult = false;

	// Required for failure cleanup code.
	if (NULL != pNameBuffer)
	{
		(*pNameBuffer) = NULL;
	}

	const char *sectionTitle = m_title;

	if (NULL != sectionTitle)
	{
		const char *devicePrefix = "Device(";
		size_t devicePrefixLength = strlen(devicePrefix);

		if (0 == _strnicmp(sectionTitle, devicePrefix, devicePrefixLength))
		{
			sectionTitle += devicePrefixLength;

			size_t tokenBufferSize = strlen(sectionTitle) + 1;

			const char *bufPtr = sectionTitle;
			char *tokenBuffer = new char[tokenBufferSize];

			// Device name is the first argument
			if (Tokenize(&bufPtr, ",", ")", tokenBuffer, tokenBufferSize, false, true, false))
			{
				if (NULL != pNameBuffer)
				{
					(*pNameBuffer) = new char[strlen(tokenBuffer) + 1];
					strcpy(*pNameBuffer, tokenBuffer);
				}

				bResult = true; // Looks like a device to me.
			}

			delete [] tokenBuffer;
		}
	}

	// Failure cleanup code.
	if (!bResult)
	{
		// If we're returning false, free partial results.

		if ((NULL != pNameBuffer) && (NULL != (*pNameBuffer)))
		{
			delete [] (*pNameBuffer);
			(*pNameBuffer) = NULL;
		}
	}

	return(bResult);
}

// Returns true if a panel was parsed and the output arguments are valid.
// If returns true and pNameBuffer not NULL on input, *pNameBuffer is the name and should be delete[]'d.
bool CCSFSection::ScanForPanelEscaped(char **pNameBuffer) const
{
	bool bResult = false;

	// Required for failure cleanup code.
	if (NULL != pNameBuffer)
	{
		(*pNameBuffer) = NULL;
	}

	const char *sectionTitle = m_title;

	if (NULL != sectionTitle)
	{
		const char *panelPrefix = "Panel(";
		size_t panelPrefixLength = strlen(panelPrefix);

		if (0 == _strnicmp(sectionTitle, panelPrefix, panelPrefixLength))
		{
			sectionTitle += panelPrefixLength;

			size_t tokenBufferSize = strlen(sectionTitle) + 1;

			const char *bufPtr = sectionTitle;
			char *tokenBuffer = new char[tokenBufferSize];

			// Panel name is the first argument
			if (Tokenize(&bufPtr, ",", ")", tokenBuffer, tokenBufferSize, false, true, false))
			{
				if (NULL != pNameBuffer)
				{
					(*pNameBuffer) = new char[strlen(tokenBuffer) + 1];
					strcpy(*pNameBuffer, tokenBuffer);
				}

				bResult = true; // Looks like a panel to me.
			}

			delete [] tokenBuffer;
		}
	}

	// Failure cleanup code.
	if (!bResult)
	{
		// If we're returning false, free partial results.

		if ((NULL != pNameBuffer) && (NULL != (*pNameBuffer)))
		{
			delete [] (*pNameBuffer);
			(*pNameBuffer) = NULL;
		}
	}

	return(bResult);
}

// Returns true if a frame was parsed.
bool CCSFSection::ScanForFrame(void) const
{
	bool bResult = false;

	if (NULL != m_title)
	{
		const char *framePrefix = "Frame(";
		size_t framePrefixLength = strlen(framePrefix);

		if (0 == _strnicmp(m_title, framePrefix, framePrefixLength))
		{
			bResult = true; // Looks like a frame to me.
		}
	}

	return(bResult);
}

// Returns true if a Device Jump was parsed and the ouput arguments are valid.
// If returns true and pDeviceNameBuffer not NULL on input, *pDeviceNameBuffer is the device name and should be delete[]'d.
// If returns true and pPanelNameBuffer not NULL on input, *pPanelNameBuffer is the device name and should be delete[]'d.
// pDeviceNameBuffer may be "this" (literally), although the function will attempt to resolve to a real device name.
bool CCSFSection::ScanForDevicePanelJumpEscaped(char **pDeviceNameBuffer, char **pPanelNameBuffer)
{
	bool bResult = false;

	// Required for failure cleanup code.
	if (NULL != pDeviceNameBuffer)
	{
		(*pDeviceNameBuffer) = NULL;
	}

	if (NULL != pPanelNameBuffer)
	{
		(*pPanelNameBuffer) = NULL;
	}

	const char *sectionTitle = m_title;

	if (NULL != sectionTitle)
	{
		const char *devicePanelJumpPrefix = "JumpDevicePanel(";
		size_t devicePanelJumpPrefixLength = strlen(devicePanelJumpPrefix);

		if (0 == _strnicmp(sectionTitle, devicePanelJumpPrefix, devicePanelJumpPrefixLength))
		{
			sectionTitle += devicePanelJumpPrefixLength;

			size_t tokenBufferSize = strlen(sectionTitle) + 1;

			const char *bufPtr = sectionTitle;
			char *tokenBuffer = new char[tokenBufferSize];

			// Device name is the first argument
			// Panel name is the second argument
			if (Tokenize(&bufPtr, ",", ")", tokenBuffer, tokenBufferSize, false, true, false))
			{
				if (NULL != pDeviceNameBuffer)
				{
					(*pDeviceNameBuffer) = new char[strlen(tokenBuffer) + 1];
					strcpy(*pDeviceNameBuffer, tokenBuffer);

					// Attempt to resolve if it's "this" (literally)
					if (0 == _stricmp("this", (*pDeviceNameBuffer)))
					{
						char *deviceName = NULL;

						for (const CCSFSection *pParent = GetParent(); NULL != pParent; pParent = pParent->GetParent())
						{
							if (pParent->ScanForDeviceEscaped(&deviceName))
							{
								delete [] (*pDeviceNameBuffer);
								(*pDeviceNameBuffer) = deviceName;
								break;
							}
						}
					}
				}

				if (Tokenize(&bufPtr, ",", ")", tokenBuffer, tokenBufferSize, false, true, false))
				{
					if (NULL != pPanelNameBuffer)
					{
						(*pPanelNameBuffer) = new char[strlen(tokenBuffer) + 1];
						strcpy(*pPanelNameBuffer, tokenBuffer);
					}
				
					bResult = true; // Looks like a device panel jump to me.
				}
			}

			delete [] tokenBuffer;
		}
	}

	// Failure cleanup code.
	if (!bResult)
	{
		// If we're returning false, free partial results.

		if ((NULL != pDeviceNameBuffer) && (NULL != (*pDeviceNameBuffer)))
		{
			delete [] (*pDeviceNameBuffer);
			(*pDeviceNameBuffer) = NULL;
		}

		if ((NULL != pPanelNameBuffer) && (NULL != (*pPanelNameBuffer)))
		{
			delete [] (*pPanelNameBuffer);
			(*pPanelNameBuffer) = NULL;
		}
	}

	return(bResult);
}

//static
bool CCSFSection::IsHardButtonName(const char *name)
{
	static const char * const hardNames[] =
	{
		"Left", "Right", "Vol-", "Vol+", "Chan-", "Chan+", "Mute", NULL
	};

	bool bResult = false;

	if (NULL != name)
	{
		for (const char * const *pHN = hardNames; NULL != *pHN; pHN++)
		{
			if (0 == (stricmp(name, *pHN)))
			{
				bResult = true;
				break;
			}
		}
	}

	return(bResult);
}

/////////////////////////////////////////////////////////////////////////
