// LayoutNudelist.cpp: implementation of the CLayoutNudelist class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "LayoutNudelist.h"
#include "LayoutNudelistLines.h"
#include "SNVolume.h"
#include "SNHardButtonDefinition.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLayoutNudelist::CLayoutNudelist(const char *versionString, const CSNVolume *pVolume, const char *deviceNameEscaped,
								 const CSNHardButtonDefinitionList *pHardButtonDefinitionList, int nPanelCap)
: m_pVolume(pVolume),
  m_strDeviceNameEscaped(deviceNameEscaped),
  m_pHardButtons(pHardButtonDefinitionList),
  m_nPanelCap(nPanelCap),
  m_pRootLayoutItem(NULL),
  m_pLettersPanel(NULL),
  m_pCreditsPanel(NULL)
{
	m_strVersionString = NULL != versionString ? versionString : "";
}

CLayoutNudelist::~CLayoutNudelist()
{
	delete m_pLettersPanel;
	m_pLettersPanel = NULL;

	delete m_pCreditsPanel;
	m_pCreditsPanel = NULL;

	while (! m_artistPanels.empty() )
	{
		delete m_artistPanels.back();
		m_artistPanels.pop_back();
	}

	while (! m_discPanels.empty() )
	{
		delete m_discPanels.back();
		m_discPanels.pop_back();
	}

	while (! m_trackPanels.empty() )
	{
		delete m_trackPanels.back();
		m_trackPanels.pop_back();
	}

	delete m_pRootLayoutItem;
	m_pRootLayoutItem = NULL;
}

const CActionSectionAdder *CLayoutNudelist::DoLayout(const CSNVolume *pVolume)
{
	delete m_pRootLayoutItem;
	m_pRootLayoutItem = NULL;

	if (pVolume->GetShowLetters())
	{
		m_pRootLayoutItem = new CLNItemLetters(this, pVolume);
	}
	else if (pVolume->GetShowArtists())
	{
		m_pRootLayoutItem = new CLNItemArtists(this, pVolume);
	}
	else if (pVolume->GetShowDiscs())
	{
		m_pRootLayoutItem = new CLNItemDiscs(this, pVolume);
	}
	else if (pVolume->GetShowTracks())
	{
		m_pRootLayoutItem = new CLNItemTracks(this, pVolume);
	}
	else
	{
		// Assert here instead of this just-in-case default?
		m_pRootLayoutItem = new CLNItemLetters(this, pVolume);
	}

	return(m_pRootLayoutItem->MakePanelsAndGetAction(NULL));
}

int CLayoutNudelist::getCurrentDeviceNumber(void)
{
	if (0 >= m_nPanelCap)
	{
		return(0);
	}
	else
	{
		int nTotal = 0;

		if (NULL != m_pLettersPanel)
		{
			nTotal++;
		}

		if (NULL != m_pCreditsPanel)
		{
			nTotal++;
		}

		nTotal += m_artistPanels.size();
		nTotal += m_discPanels.size();
		nTotal += m_trackPanels.size();

		const int nMaxPanelsPerDevice = 50;

		int nCurrentDeviceNumber = nTotal / m_nPanelCap;

		return(nCurrentDeviceNumber);
	}
}

void CLayoutNudelist::AddToDevicesToRootCSFSection(CCSFSection *pRootCSFSection) const
{
	list< CLNPanel * > allPanels;

	// Make a list of all the panels.

	if (NULL != m_pLettersPanel)
	{
		allPanels.push_back(m_pLettersPanel);
	}

	if (NULL != m_pCreditsPanel)
	{
		allPanels.push_back(m_pCreditsPanel);
	}

	list< CLNPLines * >::const_iterator iterLines;

	for (iterLines = m_artistPanels.begin(); iterLines != m_artistPanels.end(); iterLines++)
	{
		allPanels.push_back(*iterLines);
	}

	for (iterLines = m_discPanels.begin(); iterLines != m_discPanels.end(); iterLines++)
	{
		allPanels.push_back(*iterLines);
	}

	for (iterLines = m_trackPanels.begin(); iterLines != m_trackPanels.end(); iterLines++)
	{
		allPanels.push_back(*iterLines);
	}

	// Create all the required devices sorted by name so it doesn't look ugly.

	map< int, pair< string, CCSFSection *> > devMap;
	map< int, pair< string, CCSFSection *> >::iterator devMapIter;

	list< CLNPanel * >::const_iterator iterPanels;

	for (iterPanels = allPanels.begin(); iterPanels != allPanels.end(); iterPanels++)
	{
		devMapIter = devMap.find((*iterPanels)->GetDeviceNumber());

		if (devMapIter == devMap.end())
		{
			// If it's not in the map already, add in a blank entry.
			devMap.insert(pair< int, pair< string, CCSFSection *> >( (*iterPanels)->GetDeviceNumber(),
										pair< string, CCSFSection * >((*iterPanels)->GetDeviceName(), NULL) ));
		}
	}

	for (devMapIter = devMap.begin(); devMapIter != devMap.end(); devMapIter++)
	{
		// Taking advantage of the fact that maps sort by their key we now create the devices
		// in order and fill in the blanks in the map.
		devMapIter->second.second = addNewDeviceToRootCSFSection(pRootCSFSection, devMapIter->second.first.c_str());
	}

	// Add all those panels


	for (iterPanels = allPanels.begin(); iterPanels != allPanels.end(); iterPanels++)
	{
		devMapIter = devMap.find((*iterPanels)->GetDeviceNumber());

		if (devMapIter != devMap.end())
		{
			(*iterPanels)->AddToDeviceCSFSection(devMapIter->second.second);
		}
	}
}

CCSFSection *CLayoutNudelist::addNewDeviceToRootCSFSection(CCSFSection *pRootCSFSection, const char *szDeviceName) const
{
	CString deviceTitle;
	deviceTitle.Format("Device(\"%s\")", szDeviceName);

	CCSFSection *pCSFSectionListDevice = pRootCSFSection->AddSubSection();
	pCSFSectionListDevice->SetTitleEscaped(deviceTitle);

	// Add Hard buttons to list device.

	for (const CSNHardButtonDefinition *pHBD = m_pHardButtons->GetFirst(); NULL != pHBD; pHBD = m_pHardButtons->GetNext(pHBD))
	{
		if (pHBD->IsNonEmpty())
		{
			pHBD->AddToDeviceCSFSection(pCSFSectionListDevice);
		}
	}

	return(pCSFSectionListDevice);
}

//*********************************************************************************************
//*********************************************************************************************
//*********************************************************************************************

CCompositeActionSection::CCompositeActionSection(const CActionSectionAdder *pFirst, const CActionSectionAdder *pSecond)
: m_pFirst(pFirst),
  m_pSecond(pSecond)
{
}

void CCompositeActionSection::AddToActionSection(CCSFSection *pActionSection) const
{
	if (NULL != m_pFirst)
	{
		m_pFirst->AddToActionSection(pActionSection);
	}

	if (NULL != m_pSecond)
	{
		m_pSecond->AddToActionSection(pActionSection);
	}
}

//*********************************************************************************************
//*********************************************************************************************
//*********************************************************************************************

CJumpActionAdder::CJumpActionAdder(const char *szDeviceNameEscaped,
								   const char *szPanelNameEscaped)
: m_strDeviceNameEscaped(szDeviceNameEscaped),
  m_strPanelNameEscaped(szPanelNameEscaped)
{
}

CJumpActionAdder::~CJumpActionAdder()
{
}

void CJumpActionAdder::AddToActionSection(CCSFSection *pActionSection) const
{
	pActionSection->AddEscapedTitleOnlySubSectionFormat("JumpDevicePanel(\"%s\", \"%s\");",
														m_strDeviceNameEscaped.c_str(),
														m_strPanelNameEscaped.c_str());
}

//*********************************************************************************************
//*********************************************************************************************
//*********************************************************************************************

CLNItem::CLNItem(CLayoutNudelist *pLayout, const char *szTitleEscaped, const char *szShortTitleEscaped)
: m_pLayout(pLayout),
  m_strTitleEscaped(szTitleEscaped),
  m_strShortTitleEscaped(szShortTitleEscaped)
{
}

CLNItem::CLNItem(CLayoutNudelist *pLayout, char charTitleEscaped)
: m_pLayout(pLayout)
{
	m_strTitleEscaped += charTitleEscaped;
	m_strShortTitleEscaped += charTitleEscaped;
}

CLNItem::~CLNItem()
{
}

const char *CLNItem::GetTitleEscaped(void)
{
	return(m_strTitleEscaped.c_str());
}

const char *CLNItem::GetShortTitleEscaped(void)
{
	return(m_strShortTitleEscaped.c_str());
}

const CSNVolume *CLNItem::getVolume(void)
{
	return(m_pLayout->m_pVolume);
}

const char *CLNItem::getDeviceNameEscaped(void)
{
	return(m_pLayout->m_strDeviceNameEscaped.c_str());
}

const char *CLNItem::getVersionString(void)
{
	return(m_pLayout->m_strVersionString.c_str());
}

CLNAlphabetPanel ** CLNItem::getLettersPanelPointer(void)
{
	return(&m_pLayout->m_pLettersPanel);
}

CLNCreditsPanel ** CLNItem::getCreditsPanelPointer(void)
{
	return(&m_pLayout->m_pCreditsPanel);
}

list<CLNPLines *> & CLNItem::getArtistPanels(void)
{
	return(m_pLayout->m_artistPanels);
}

list<CLNPLines *> & CLNItem::getDiscPanels(void)
{
	return(m_pLayout->m_discPanels);
}

list<CLNPLines *> & CLNItem::getTrackPanels(void)
{
	return(m_pLayout->m_trackPanels);
}

int CLNItem::getCurrentDeviceNumber(void)
{
	return(m_pLayout->getCurrentDeviceNumber());
}

//*********************************************************************************************

CLNItemWithChildren::CLNItemWithChildren(CLayoutNudelist *pLayout, const char *szTitleEscaped, const char *szShortTitleEscaped)
: CLNItem(pLayout, szTitleEscaped, szShortTitleEscaped),
  m_intCachedValidActionResult(0)
{
}

CLNItemWithChildren::CLNItemWithChildren(CLayoutNudelist *pLayout, char charTitleEscaped)
: CLNItem(pLayout, charTitleEscaped),
  m_intCachedValidActionResult(0)
{
}

CLNItemWithChildren::~CLNItemWithChildren()
{
	while (! m_children.empty() )
	{
		delete m_children.back();
		m_children.pop_back();
	}
}

// Predicts whether the return of MakePanelsAndGetAction will be NULL or not.
bool CLNItemWithChildren::IsGoingToReturnValidAction(void)
{
	bool bResult = false;

	if (0 < m_intCachedValidActionResult)
	{
		bResult = true;
	}
	else if (0 == m_intCachedValidActionResult)
	{
		for (list<CLNItem *>::iterator chIter = m_children.begin(); chIter != m_children.end(); chIter++)
		{
			if ((*chIter)->IsGoingToReturnValidAction())
			{
				bResult = true;
				break;
			}
		}

		if (bResult)
		{
			m_intCachedValidActionResult = 1;
		}
		else
		{
			m_intCachedValidActionResult = -1;
		}
	}

	return(bResult);
}

// Returns NULL if the child list is empty and there is no appropriate default action.
// Whether or not the return value is NULL can be predicted by calling IsGoingToReturnValidAction
// pParentPanel can be NULL if there isn't one.
// If a panel jump is returned it will be absolute iff pParentPanel is NULL.
const CActionSectionAdder *CLNItemWithChildren::MakePanelsAndGetAction(CLNPanel *pParentPanel)
{
	const CActionSectionAdder *pASAResult = NULL;

	list<CLNItem *>::iterator chIter;

	// Calculate the total number of children which will return valid action sections.
	// Those which return invalid action sections are empty or void or whatever.

	unsigned int numValidChildren = 0;

	for (chIter = m_children.begin(); chIter != m_children.end(); chIter++)
	{
		if ((*chIter)->IsGoingToReturnValidAction())
		{
			numValidChildren++;
		}
	}

	if (0 < numValidChildren)
	{
		list<CLNPLines *> & targetPanelList = getTargetPanel();

		// New panel for this collection ?

		bool bAddSpaceBeforeFirstHeading = false;

		if (targetPanelList.empty())
		{
			// Create first panel if target list is empty.
			targetPanelList.push_back(new CLNPLines(getVolume(), getDeviceNameEscaped(), getCurrentDeviceNumber(), getPanelNamePrefix(), 0));

			targetPanelList.back()->SetLeftActionAndMsgEscaped(NULL, "[start]", 0);
			targetPanelList.back()->SetRightActionAndMsgEscaped(NULL, "[end]", 0);
		}
		else
		{
			bool bAddPanel = false;

			if (getVolume()->GetGapBreakAlways() && (!targetPanelList.back()->IsEmpty()))
			{
				bAddPanel = true;
			}
			else
			{
				int reqLines = 2;

				if (getVolume()->GetGapSpace())
				{
					reqLines++;
				}

				if (getVolume()->GetGapFit() && (2 < numValidChildren))
				{
					// (n-1)/2 means it calculates enough lines when there is
					// an odd number of children remaining. For example:
					// 8 remain, 6 unallocated: (8-1)/2 = 3 which can hold the 6.
					// 7 remain, 5 unallocated: (7-1)/2 = 3 which can hold the 5.
					// (It is really ((n-2)+1)/2) .)
					reqLines += ((numValidChildren - 1) / 2);
				}

				if (reqLines > targetPanelList.back()->FreeLines())
				{
					bAddPanel = true;
				}
				else if (getVolume()->GetGapSpace())
				{
					bAddSpaceBeforeFirstHeading = true;
				}
			}

			if (bAddPanel)
			{
				// Create new panel if last panel in target list is not empty.

				CLNPLines *pLastPanel = targetPanelList.back();

				targetPanelList.push_back(new CLNPLines(getVolume(), getDeviceNameEscaped(), getCurrentDeviceNumber(), getPanelNamePrefix(), targetPanelList.back()->GetPanelNumber() + 1));

				pLastPanel->SetRightActionAndMsgEscaped(targetPanelList.back()->GetJumpActionAdder(false), this->GetShortTitleEscaped(), 0);
				targetPanelList.back()->SetLeftActionAndMsgEscaped(pLastPanel->GetJumpActionAdder(false), pLastPanel->GetLastShortHeadingEscaped(), 0);
				targetPanelList.back()->SetRightActionAndMsgEscaped(NULL, "[end]", 0);
			}
		}

		pASAResult = targetPanelList.back()->GetJumpActionAdder(NULL == pParentPanel);
		pASAResult = getAugmentedActionSectionAdder(pASAResult);

		// Header for this collection.
		{
			const CActionSectionAdder *pParentJump = NULL;

			if (NULL != pParentPanel)
			{
				pParentJump = pParentPanel->GetJumpActionAdder(false);
			}

			if (bAddSpaceBeforeFirstHeading)
			{
				targetPanelList.back()->AddPanelLine(new CLNLBlank());
			}

			targetPanelList.back()->AddPanelLine(new CLNLHeading(this->GetTitleEscaped(), pParentJump));
			targetPanelList.back()->SetLastShortHeadingEscaped(this->GetShortTitleEscaped());
		}

		// Add children

		unsigned int numChildrenDone = 0;

		chIter = m_children.begin();

		while(numChildrenDone < numValidChildren)
		{
			CLNItem *pChildOne;

			do
			{
				pChildOne = *chIter++;
			} while (! pChildOne->IsGoingToReturnValidAction() );

			numChildrenDone++;

			const CActionSectionAdder *pChildASAOne = pChildOne->MakePanelsAndGetAction(targetPanelList.back());

			if (((numChildrenDone + 1) < numValidChildren)
			&&	(1 == targetPanelList.back()->FreeLines()))
			{
				// We need a continuation message and new panel if the current panel has only
				// one line left and there are more than two children left.

				CLNPLines *pPanelThatRanOutOfSpace = targetPanelList.back();

				// Create and add new panel.
				targetPanelList.push_back(new CLNPLines(getVolume(), getDeviceNameEscaped(), getCurrentDeviceNumber(), getPanelNamePrefix(), targetPanelList.back()->GetPanelNumber() + 1));

				// Add current item and continuation line to panel which ran out of space.
				pPanelThatRanOutOfSpace->AddPanelLine(
					new CLNLOneItemAndContinuation(pChildOne->GetTitleEscaped(), pChildOne->GetButtonLabelEscaped(), pChildASAOne,
										targetPanelList.back()->GetJumpActionAdder(false)));

				// Continuation header in new panel.
				targetPanelList.back()->AddPanelLine(
					new CLNLContinuedHeading(this->GetTitleEscaped(),
										pPanelThatRanOutOfSpace->GetJumpActionAdder(false)));

				targetPanelList.back()->SetLastShortHeadingEscaped(this->GetShortTitleEscaped());

				pPanelThatRanOutOfSpace->SetRightActionAndMsgEscaped(targetPanelList.back()->GetJumpActionAdder(false), this->GetShortTitleEscaped(), numValidChildren - numChildrenDone);
				targetPanelList.back()->SetLeftActionAndMsgEscaped(pPanelThatRanOutOfSpace->GetJumpActionAdder(false), this->GetShortTitleEscaped(), numChildrenDone);
				targetPanelList.back()->SetRightActionAndMsgEscaped(NULL, "[end]", 0);
			}
			else
			{
				if (numChildrenDone < numValidChildren)
				{
					CLNItem *pChildTwo;

					do
					{
						pChildTwo = *chIter++;
					} while (! pChildTwo->IsGoingToReturnValidAction() );

					numChildrenDone++;

					const CActionSectionAdder *pChildASATwo = pChildTwo->MakePanelsAndGetAction(targetPanelList.back());

					// Add current two items in a line.
					targetPanelList.back()->AddPanelLine(
						new CLNLTwoItems(pChildOne->GetTitleEscaped(), pChildOne->GetButtonLabelEscaped(), pChildASAOne,
										 pChildTwo->GetTitleEscaped(), pChildTwo->GetButtonLabelEscaped(), pChildASATwo));
				}
				else
				{
					// Add current and last item in a line on its own.
					targetPanelList.back()->AddPanelLine(
						new CLNLTwoItems(pChildOne->GetTitleEscaped(), pChildOne->GetButtonLabelEscaped(), pChildASAOne,
										 NULL, NULL, NULL));
				}
			}
		}
	}

	return(pASAResult);
}

//*********************************************************************************************
//*********************************************************************************************
//*********************************************************************************************

CLNItemLetters::CLNItemLetters(CLayoutNudelist *pLayout, const CSNVolume *pVolume)
: CLNItem(pLayout, pVolume->GetName()->GetEscaped(true), pVolume->GetName()->GetEscaped(true)),
  m_intCachedValidActionResult(0)
{
	char *letterArray = "#ABCDEFGHIJKLMNOPQRSTUVWXYZ";

	for (char *p = letterArray; '\0' != *p; p++)
	{
		if (getVolume()->GetShowArtists())
		{
			m_children.push_back(new CLNItemArtists(pLayout, pVolume, *p));
		}
		else if (getVolume()->GetShowDiscs())
		{
			m_children.push_back(new CLNItemDiscs(pLayout, pVolume, *p));
		}
		else if (getVolume()->GetShowTracks())
		{
			m_children.push_back(new CLNItemTracks(pLayout, pVolume, *p));
		}
	}
}

CLNItemLetters::~CLNItemLetters()
{
	while (! m_children.empty() )
	{
		delete m_children.back();
		m_children.pop_back();
	}
}

// Predicts whether the return of MakePanelsAndGetAction will be NULL or not.
bool CLNItemLetters::IsGoingToReturnValidAction(void)
{
	bool bResult = false;

	if (0 < m_intCachedValidActionResult)
	{
		bResult = true;
	}
	else if (0 == m_intCachedValidActionResult)
	{
		for (list<CLNItem *>::iterator chIter = m_children.begin(); chIter != m_children.end(); chIter++)
		{
			if ((*chIter)->IsGoingToReturnValidAction())
			{
				bResult = true;
				break;
			}
		}

		if (bResult)
		{
			m_intCachedValidActionResult = 1;
		}
		else
		{
			m_intCachedValidActionResult = -1;
		}
	}

	return(bResult);
}

// Returns NULL if the child list is empty and there is no appropriate default action.
// Whether or not the return value is NULL can be predicted by calling IsGoingToReturnValidAction
// pParentPanel can be NULL if there isn't one.
// If a panel jump is returned it will be absolute iff pParentPanel is NULL.
const CActionSectionAdder *CLNItemLetters::MakePanelsAndGetAction(CLNPanel *pParentPanel)
{
	const CActionSectionAdder *pASAResult = NULL;

	list<CLNItem *>::iterator chIter;

	// Calculate the total number of children which will return valid action sections.
	// Those which return invalid action sections are empty or void or whatever.

	unsigned int numValidChildren = 0;

	for (chIter = m_children.begin(); chIter != m_children.end(); chIter++)
	{
		if ((*chIter)->IsGoingToReturnValidAction())
		{
			numValidChildren++;
		}
	}

	CLNAlphabetPanel **ppLettersPanel = getLettersPanelPointer();

	if ((0 < numValidChildren) && (NULL == *ppLettersPanel))
	{
		// New panel for this alphabet.
		(*ppLettersPanel) = new CLNAlphabetPanel(getVolume(), getDeviceNameEscaped(), getCurrentDeviceNumber());

		pASAResult = (*ppLettersPanel)->GetJumpActionAdder(NULL == pParentPanel);

		CLNCreditsPanel **ppCreditsPanel = getCreditsPanelPointer();

		if (NULL == *ppCreditsPanel)
		{
			(*ppCreditsPanel) = new CLNCreditsPanel(getVolume(), getDeviceNameEscaped(), getCurrentDeviceNumber(), *ppLettersPanel);

			(*ppCreditsPanel)->SetLeftActionAndMsgEscaped((*ppLettersPanel)->GetJumpActionAdder(false), "Alphabet", 0);
			(*ppCreditsPanel)->SetRightActionAndMsgEscaped((*ppLettersPanel)->GetJumpActionAdder(false), "Alphabet", 0);
		}

		(*ppLettersPanel)->SetLeftActionAndMsgEscaped((*ppCreditsPanel)->GetJumpActionAdder(false), "Super-Nudelist", 0);
		(*ppLettersPanel)->SetRightActionAndMsgEscaped((*ppCreditsPanel)->GetJumpActionAdder(false), getVersionString(), 0);

		// Add children

		unsigned int numChildrenDone = 0;

		chIter = m_children.begin();

		while(numChildrenDone < numValidChildren)
		{
			CLNItem *pChildOne;

			do
			{
				pChildOne = *chIter++;
			} while (! pChildOne->IsGoingToReturnValidAction() );

			numChildrenDone++;

			const CActionSectionAdder *pChildASAOne = \
					pChildOne->MakePanelsAndGetAction(*ppLettersPanel);

			(*ppLettersPanel)->NotePanelAction(GetAlphabetPanelLetterGroup(pChildOne->GetTitleEscaped()), pChildASAOne);
		}
	}

	return(pASAResult);
}

//*********************************************************************************************
//*********************************************************************************************
//*********************************************************************************************

template<> struct std::greater<const CSNArtist *>
{
	bool operator()(const CSNArtist*& left, const CSNArtist*& right) const
	{
		return(0 > left->CompareForSort(right));
	}
};

CLNItemArtists::CLNItemArtists(CLayoutNudelist *pLayout, const CSNVolume *pVolume, char letterEscaped)
: CLNItemWithChildren(pLayout, letterEscaped)
{
	char normLetter = GetAlphabetPanelLetterGroup(letterEscaped);

	list<const CSNArtist *> queuedArtists;

	for (const CSNArtist *pArtist = pVolume->GetFirst(); NULL != pArtist; pArtist = pVolume->GetNext(pArtist))
	{
		if (normLetter == GetAlphabetPanelLetterGroup(pArtist->GetSortAsNameOrName()->GetPronto(false)))
		{
			queuedArtists.push_back(pArtist);
		}
	}

	queuedArtists.sort(greater<const CSNArtist *>());

	while (!queuedArtists.empty())
	{
		addArtist(pLayout, queuedArtists.front());
		queuedArtists.pop_front();
	}
}

CLNItemArtists::CLNItemArtists(CLayoutNudelist *pLayout, const CSNVolume *pVolume)
: CLNItemWithChildren(pLayout, pVolume->GetName()->GetEscaped(true), pVolume->GetName()->GetEscaped(true))
{
	for (const CSNArtist *pArtist = pVolume->GetFirst(); NULL != pArtist; pArtist = pVolume->GetNext(pArtist))
	{
		addArtist(pLayout, pArtist);
	}
}

void CLNItemArtists::addArtist(CLayoutNudelist *pLayout, const CSNArtist *pArtist)
{
	if (getVolume()->GetShowDiscs())
	{
		m_children.push_back(new CLNItemDiscs(pLayout, pArtist));
	}
	else if (getVolume()->GetShowTracks())
	{
		m_children.push_back(new CLNItemTracks(pLayout, pArtist));
	}
}

//*********************************************************************************************
//*********************************************************************************************
//*********************************************************************************************

template<> struct std::greater<const CSNDisc *>
{
	bool operator()(const CSNDisc*& left, const CSNDisc*& right) const
	{
		return(0 > left->CompareForSort(right));
	}
};

CLNItemDiscs::CLNItemDiscs(CLayoutNudelist *pLayout, const CSNVolume *pVolume, char letterEscaped)
: CLNItemWithChildren(pLayout, letterEscaped)
{
	char normLetter = GetAlphabetPanelLetterGroup(letterEscaped);

	list<const CSNDisc *> queuedDiscs;

	for (const CSNArtist *pArtist = pVolume->GetFirst(); NULL != pArtist; pArtist = pVolume->GetNext(pArtist))
	{
		for (const CSNDisc *pDisc = pArtist->GetFirst(); NULL != pDisc; pDisc = pArtist->GetNext(pDisc))
		{
			if (normLetter == GetAlphabetPanelLetterGroup(pDisc->GetSortAsNameOrName()->GetPronto(false)))
			{
				queuedDiscs.push_back(pDisc);
			}
		}
	}

	queuedDiscs.sort(greater<const CSNDisc *>());

	while (!queuedDiscs.empty())
	{
		addDisc(pLayout, queuedDiscs.front());
		queuedDiscs.pop_front();
	}
}

CLNItemDiscs::CLNItemDiscs(CLayoutNudelist *pLayout, const CSNVolume *pVolume)
: CLNItemWithChildren(pLayout, pVolume->GetName()->GetEscaped(true), pVolume->GetName()->GetEscaped(true))
{
	for (const CSNArtist *pArtist = pVolume->GetFirst(); NULL != pArtist; pArtist = pVolume->GetNext(pArtist))
	{
		for (const CSNDisc *pDisc = pArtist->GetFirst(); NULL != pDisc; pDisc = pArtist->GetNext(pDisc))
		{
			addDisc(pLayout, pDisc);
		}
	}
}

CLNItemDiscs::CLNItemDiscs(CLayoutNudelist *pLayout, const CSNArtist *pArtist)
: CLNItemWithChildren(pLayout, pArtist->GetName()->GetEscaped(true), pArtist->GetShortNameOrName()->GetEscaped(true))
{
	for (const CSNDisc *pDisc = pArtist->GetFirst(); NULL != pDisc; pDisc = pArtist->GetNext(pDisc))
	{
		addDisc(pLayout, pDisc);
	}
}

void CLNItemDiscs::addDisc(CLayoutNudelist *pLayout, const CSNDisc *pDisc)
{
	if ((getVolume()->GetShowTracks())
	&&	(! pDisc->GetIgnoreTracks() )
	&&	(0 != pDisc->GetTracks()->GetNumElements() ))
	{
		m_children.push_back(new CLNItemTracks(pLayout, pDisc));
	}
	else
	{
		m_children.push_back(new CLNItemActionDisc(pLayout, pDisc, true));
	}
}

//*********************************************************************************************
//*********************************************************************************************
//*********************************************************************************************

CLNItemActionDisc::CLNItemActionDisc(CLayoutNudelist *pLayout, const CSNDisc *pDisc, bool bIncludeJump)
: CLNItem(pLayout, pDisc->GetName()->GetEscaped(true), pDisc->GetShortNameOrName()->GetEscaped(true)),
  m_bIncludeJump(bIncludeJump),
  m_pDisc(pDisc),
  m_szButtonLabelCache(NULL)
{
}

CLNItemActionDisc::~CLNItemActionDisc()
{
	delete [] m_szButtonLabelCache;
	m_szButtonLabelCache = NULL;
}

// override of non-pure virtual
const char *CLNItemActionDisc::GetButtonLabelEscaped(void)
{
	if (NULL != m_szButtonLabelCache)
	{
		return(m_szButtonLabelCache);
	}
	else if (getVolume()->GetShowNumbers())
	{
		CString csTemp = m_pDisc->GetDiscNumberLabelEscaped();

		m_szButtonLabelCache = new char[csTemp.GetLength() + 1];
		strcpy(m_szButtonLabelCache, LPCTSTR(csTemp));

		return(m_szButtonLabelCache);
	}
	else
	{
		return("");
	}
}

bool CLNItemActionDisc::IsGoingToReturnValidAction(void)
{
	return(true);
}

const CActionSectionAdder *CLNItemActionDisc::MakePanelsAndGetAction(CLNPanel *pParentPanel)
{
	return(this);
}

void CLNItemActionDisc::AddToActionSection(CCSFSection *pActionSection) const
{
	m_pDisc->AddDiscCallToButtonAction(pActionSection, m_bIncludeJump);
}

//*********************************************************************************************
//*********************************************************************************************
//*********************************************************************************************

template<> struct std::greater<const CSNTrack *>
{
	bool operator()(const CSNTrack*& left, const CSNTrack*& right) const
	{
		return(0 > left->GetTitle()->Compare(right->GetTitle(), false));
	}
};

CLNItemTracks::CLNItemTracks(CLayoutNudelist *pLayout, const CSNVolume *pVolume, char letterEscaped)
: CLNItemWithChildren(pLayout, letterEscaped),
  m_pDisc(NULL),
  m_pLayout(pLayout),
  m_szButtonLabelCache(NULL)
{
	char normLetter = GetAlphabetPanelLetterGroup(letterEscaped);

	list<const CSNTrack *> queuedTracks;

	for (const CSNArtist *pArtist = pVolume->GetFirst(); NULL != pArtist; pArtist = pVolume->GetNext(pArtist))
	{
		for (const CSNDisc *pDisc = pArtist->GetFirst(); NULL != pDisc; pDisc = pArtist->GetNext(pDisc))
		{
			for (const CSNTrack *pTrack = pDisc->GetTracks()->GetFirst(); NULL != pTrack; pTrack = pDisc->GetTracks()->GetNext(pTrack))
			{
				if (normLetter == GetAlphabetPanelLetterGroup(pTrack->GetTitle()->GetPronto(false)))
				{
					queuedTracks.push_back(pTrack);
				}
			}
		}
	}

	queuedTracks.sort(greater<const CSNTrack *>());

	while (!queuedTracks.empty())
	{
		addTrack(pLayout, queuedTracks.front());
		queuedTracks.pop_front();
	}
}

CLNItemTracks::CLNItemTracks(CLayoutNudelist *pLayout, const CSNVolume *pVolume)
: CLNItemWithChildren(pLayout, pVolume->GetName()->GetEscaped(true), pVolume->GetName()->GetEscaped(true)),
  m_pDisc(NULL),
  m_pLayout(pLayout),
  m_szButtonLabelCache(NULL)
{
	for (const CSNArtist *pArtist = pVolume->GetFirst(); NULL != pArtist; pArtist = pVolume->GetNext(pArtist))
	{
		for (const CSNDisc *pDisc = pArtist->GetFirst(); NULL != pDisc; pDisc = pArtist->GetNext(pDisc))
		{
			for (const CSNTrack *pTrack = pDisc->GetTracks()->GetFirst(); NULL != pTrack; pTrack = pDisc->GetTracks()->GetNext(pTrack))
			{
				addTrack(pLayout, pTrack);
			}
		}
	}
}

CLNItemTracks::CLNItemTracks(CLayoutNudelist *pLayout, const CSNArtist *pArtist)
: CLNItemWithChildren(pLayout, pArtist->GetName()->GetEscaped(true), pArtist->GetShortNameOrName()->GetEscaped(true)),
  m_pDisc(NULL),
  m_pLayout(pLayout),
  m_szButtonLabelCache(NULL)
{
	for (const CSNDisc *pDisc = pArtist->GetFirst(); NULL != pDisc; pDisc = pArtist->GetNext(pDisc))
	{
		for (const CSNTrack *pTrack = pDisc->GetTracks()->GetFirst(); NULL != pTrack; pTrack = pDisc->GetTracks()->GetNext(pTrack))
		{
			addTrack(pLayout, pTrack);
		}
	}
}

CLNItemTracks::CLNItemTracks(CLayoutNudelist *pLayout, const CSNDisc *pDisc)
: CLNItemWithChildren(pLayout, pDisc->GetName()->GetEscaped(true), pDisc->GetShortNameOrName()->GetEscaped(true)),
  m_pDisc(pDisc),
  m_pLayout(pLayout),
  m_szButtonLabelCache(NULL)
{
	for (const CSNTrack *pTrack = pDisc->GetTracks()->GetFirst(); NULL != pTrack; pTrack = pDisc->GetTracks()->GetNext(pTrack))
	{
		addTrack(pLayout, pTrack);
	}
}

CLNItemTracks::~CLNItemTracks()
{
	while (! m_actionAddersToDelete.empty() )
	{
		delete m_actionAddersToDelete.front();
		m_actionAddersToDelete.pop_front();
	}

	delete [] m_szButtonLabelCache;
	m_szButtonLabelCache = NULL;
}

// override of non-pure virtual
const char *CLNItemTracks::GetButtonLabelEscaped(void)
{
	if (NULL != m_szButtonLabelCache)
	{
		return(m_szButtonLabelCache);
	}
	else if (getVolume()->GetShowDiscs() && getVolume()->GetShowNumbers() && (NULL != m_pDisc))
	{
		CString csTemp = m_pDisc->GetDiscNumberLabelEscaped();

		m_szButtonLabelCache = new char[csTemp.GetLength() + 1];
		strcpy(m_szButtonLabelCache, LPCTSTR(csTemp));

		return(m_szButtonLabelCache);
	}
	else
	{
		return("");
	}
}

const CActionSectionAdder *CLNItemTracks::getAugmentedActionSectionAdder(const CActionSectionAdder *pPanelJumpASA)
{
	// If m_pDisc is non-NULL, we need to add the disc command before the panel jump.

	if ((!getVolume()->GetShowDiscs()) || getVolume()->GetDelayDiscUntilTrack()
	||	(NULL == m_pDisc) || (NULL == m_pLayout))
	{
		return(pPanelJumpASA);
	}
	else
	{
		// Disc action without jump.
		CActionSectionAdder *pDiscASA = new CLNItemActionDisc(m_pLayout, m_pDisc, false); 

		// Remember to delete it later.
		m_actionAddersToDelete.push_back(pDiscASA);

		CActionSectionAdder *pCompositeASA = new CCompositeActionSection(pDiscASA, pPanelJumpASA);

		// Remember to delete it later.
		m_actionAddersToDelete.push_back(pCompositeASA);

		return(pCompositeASA);
	}
}

void CLNItemTracks::addTrack(CLayoutNudelist *pLayout, const CSNTrack *pTrack)
{
	// Need to give the disc to the action if there is no disc above us or if
	// disc commands are delayed until track selection.

	bool bIncludeDisc = false;

	if ((! getVolume()->GetShowDiscs() )
	||	(getVolume()->GetDelayDiscUntilTrack()))
	{
		bIncludeDisc = true;
	}

	bool bIncludeJump = getVolume()->GetJumpAfterTrack();

	m_children.push_back(new CLNItemActionTrack(pLayout, pTrack, bIncludeDisc, bIncludeJump));
}

//*********************************************************************************************
//*********************************************************************************************
//*********************************************************************************************

CLNItemActionTrack::CLNItemActionTrack(CLayoutNudelist *pLayout, const CSNTrack *pTrack, bool bIncludeDisc, bool bIncludeJump)
: CLNItem(pLayout, pTrack->GetTitle()->GetEscaped(true), pTrack->GetTitle()->GetEscaped(true)),
  m_bIncludeDisc(bIncludeDisc),
  m_bIncludeJump(bIncludeJump),
  m_pTrack(pTrack),
  m_szButtonLabelCache(NULL)
{
}

CLNItemActionTrack::~CLNItemActionTrack()
{
	delete [] m_szButtonLabelCache;
	m_szButtonLabelCache = NULL;
}

// override of non-pure virtual
const char *CLNItemActionTrack::GetButtonLabelEscaped(void)
{
	if (NULL != m_szButtonLabelCache)
	{
		return(m_szButtonLabelCache);
	}
	else if (getVolume()->GetShowNumbers())
	{
		CString csTemp = m_pTrack->GetTrackNumberLabelEscaped();

		m_szButtonLabelCache = new char[csTemp.GetLength() + 1];
		strcpy(m_szButtonLabelCache, LPCTSTR(csTemp));

		return(m_szButtonLabelCache);
	}
	else
	{
		return("");
	}
}

bool CLNItemActionTrack::IsGoingToReturnValidAction(void)
{
	return(true);
}

const CActionSectionAdder *CLNItemActionTrack::MakePanelsAndGetAction(CLNPanel *pParentPanel)
{
	return(this);
}

void CLNItemActionTrack::AddToActionSection(CCSFSection *pActionSection) const
{
	m_pTrack->AddTrackCallToButtonAction(pActionSection, m_bIncludeDisc, m_bIncludeJump);
}

//*********************************************************************************************
//*********************************************************************************************
//*********************************************************************************************
