
#include "stdafx.h"
#include <io.h>
#include "SNConfig.h"
#include "ProgressDlg.h"
#include "UnsignedBaseTenNumber.h"
#include "SNDevice.h"
#include "SNLibrary.h"

CSNConfig::CSNConfig(void)
{
	m_bConfigAndTempPathIsDefault = false;

	m_configAndTempPath = NULL;
	m_configPath = NULL;
	m_devicesFilePath = NULL;
	m_libraryFilePath = NULL;
	m_tempPath = NULL;
	m_tempCSFPath = NULL;
	m_inputCCFPath = NULL;
	m_outputCCFPath = NULL;
	m_bCompressImages = false;
	m_nPanelCap = 0;

	m_exePath = NULL;
	m_exeDirectoryPath = NULL;
	m_exeVersionString = NULL;

	m_pDeviceList = new CSNDeviceList();
	m_pLibrary = new CSNLibrary();

	InitializeCriticalSection(&m_CS);
}

CSNConfig::~CSNConfig(void)
{
	SetConfigAndTempAreaPath(NULL);
	SetInputCCFPath(NULL);
	SetOutputCCFPath(NULL);

	delete [] m_exePath;
	m_exePath = NULL;

	delete [] m_exeDirectoryPath;
	m_exeDirectoryPath = NULL;

	delete [] m_exeVersionString;
	m_exeVersionString = NULL;

	delete m_pDeviceList;

	delete m_pLibrary;

	DeleteCriticalSection(&m_CS);
}

bool CSNConfig::IsReady(CTypedError *pTError, bool bRequireOutput) const
{
	CriticalSectionScoper css(&m_CS);

	if ((NULL == m_configAndTempPath) || ('\0' == m_configAndTempPath[0]))
	{
		CTypedError::SafeSetMsgError(pTError, "Config/Temp Area undefined");
		return(false);
	}

	if ((NULL == m_inputCCFPath) || ('\0' == m_inputCCFPath[0]))
	{
		CTypedError::SafeSetMsgError(pTError, "Input CCF file not specified");
		return(false);
	}

	if (!NudelAccess(pTError, m_inputCCFPath, false))
	{
		return(false);
	}

	if (bRequireOutput)
	{		
		if ((NULL == m_outputCCFPath) || ('\0' == m_outputCCFPath[0]))
		{
			CTypedError::SafeSetMsgError(pTError, "Output CCF file not specified");
			return(false);
		}

		char *outputParent = GetParentPath(m_outputCCFPath);
		bool bExists = NudelAccess(pTError, outputParent, true);
		delete [] outputParent;
		if (!bExists)
		{
			return(false);
		}
	}

	return(true);
}

bool CSNConfig::ResolveCSFPaths(CTypedError *pTError, const CCSFAbstractTree *pAT)
{
	if (!m_pDeviceList->ResolveCSFPaths(pTError, pAT))
	{
		return(false);
	}

	if (!m_pLibrary->ResolveCSFPaths(pTError, pAT))
	{
		return(false);
	}

	return(true);
}

// Returns the set of discs which could not be resolved or NULL if there are none.
// If non-NULL, the result should be deleted.
set<CSNDisc *> *CSNConfig::ResolveDevices(void)
{
	set<CSNDisc *> *pDiscSet = new set<CSNDisc *>;

	m_pLibrary->ResolveDevices(pDiscSet, m_pDeviceList);

	if (pDiscSet->empty())
	{
		delete pDiscSet;
		pDiscSet = NULL;
	}

	return(pDiscSet);
}

bool CSNConfig::Load(CProgressDlg *pProgressDlg, CTypedError *pTError, CStringList *pSLErrorDetails)
{
	CriticalSectionScoper css(&m_CS);

	bool bResult = false;

	// -- Load path details from user's registry.

	CProgressDlg::SafeSetMessage(pProgressDlg, "Discovering paths");

	DWORD dwType;
	char *regString;

	// Config and temp area path.
	if (LeetRegQueryValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_CONFIGANDTEMPAREA,
							&dwType, (void **)(&regString), NULL))
	{
		if (REG_SZ == dwType)
		{
			SetConfigAndTempAreaPath(regString);
		}

		delete [] regString;
	}

	if (NULL == m_configAndTempPath)
	{
		SetConfigAndTempAreaPathDefault(pTError);
	}

	if (NULL != m_configAndTempPath)
	{
		bResult = true;

		CProgressDlg::SafeSetMessage(pProgressDlg, "Loading CCF paths");

		if (!LoadCCFPathsAndOptions(pTError))
		{
			bResult = false;
		}

		CProgressDlg::SafeSetMessage(pProgressDlg, "Loading Configuration");

		if (!LoadDevices(pTError, pSLErrorDetails))
		{
			bResult = false;
		}

		if (!LoadLibrary(pTError, pSLErrorDetails, NULL))
		{
			bResult = false;
		}

		CProgressDlg::SafeSetMessage(pProgressDlg, "");
	}

	return(bResult);
}

bool CSNConfig::LoadCCFPathsAndOptions(CTypedError *pTError)
{
	bool bResult = true;

	DWORD dwType;
	char *regString;

	// Input CCF path.
	if (LeetRegQueryValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_INPUTCCF,
							&dwType, (void **)(&regString), NULL))
	{
		if (REG_SZ == dwType)
		{
			SetInputCCFPath(regString);
		}

		delete [] regString;
	}

	// Output CCF path.
	if (LeetRegQueryValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_OUTPUTCCF,
							&dwType, (void **)(&regString), NULL))
	{
		if (REG_SZ == dwType)
		{
			SetOutputCCFPath(regString);
		}

		delete [] regString;
	}

	DWORD dwbCompressImages = 0;

	// Compress Images
	if (LeetRegQueryDWORDValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_COMPRESSIMAGES,
								&dwbCompressImages))
	{
		SetCompressImages(dwbCompressImages ? true : false);
	}
	else
	{
		SetCompressImages(false);
	}

	DWORD dwPanelCap = 0;

	if (LeetRegQueryDWORDValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_PANELCAP, &dwPanelCap))
	{
		SetPanelCap(dwPanelCap);
	}
	else
	{
		SetPanelCap(0);
	}

	return(bResult);
}

bool CSNConfig::SaveCCFPathsAndOptions(CTypedError *pTError) const
{
	bool bResult = true;

	CriticalSectionScoper css(&m_CS);

/*	ConfigAndTempPath registry setting is set by the user and not configurable within
	the program, so it makes sense that we should leave it how we found it.

	if ((!m_bConfigAndTempPathIsDefault) && (NULL != m_configAndTempPath))
	{
		if (! LeetRegSetValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_CONFIGANDTEMPAREA,
								REG_SZ, m_configAndTempPath, strlen(m_configAndTempPath) + 1) )
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Failed to save HKCU\\%s\\%s", SN_REG_PATH_ROOT, SN_REG_VALUE_CONFIGANDTEMPAREA);
		}
	}
	else
	{
		if (! LeetRegDeleteValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_CONFIGANDTEMPAREA) )
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Failed to delete HKCU\\%s\\%s", SN_REG_PATH_ROOT, SN_REG_VALUE_CONFIGANDTEMPAREA);
		}
	}
*/
	if (NULL != m_inputCCFPath)
	{
		if (! LeetRegSetValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_INPUTCCF,
								REG_SZ, m_inputCCFPath, strlen(m_inputCCFPath) + 1) )
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Failed to save HKCU\\%s\\%s", SN_REG_PATH_ROOT, SN_REG_VALUE_INPUTCCF);
		}
	}
	else
	{
		if (! LeetRegDeleteValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_INPUTCCF) )
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Failed to delete HKCU\\%s\\%s", SN_REG_PATH_ROOT, SN_REG_VALUE_INPUTCCF);
		}
	}

	if (NULL != m_outputCCFPath)
	{
		if (! LeetRegSetValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_OUTPUTCCF,
								REG_SZ, m_outputCCFPath, strlen(m_outputCCFPath) + 1) )
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Failed to save HKCU\\%s\\%s", SN_REG_PATH_ROOT, SN_REG_VALUE_OUTPUTCCF);
		}
	}
	else
	{
		if (! LeetRegDeleteValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_OUTPUTCCF) )
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Failed to delete HKCU\\%s\\%s", SN_REG_PATH_ROOT, SN_REG_VALUE_OUTPUTCCF);
		}
	}

	if (m_bCompressImages)
	{
		DWORD dw = 1;

		if (! LeetRegSetValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_COMPRESSIMAGES,
								REG_DWORD, &dw, sizeof(DWORD)) )
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Failed to save HKCU\\%s\\%s", SN_REG_PATH_ROOT, SN_REG_VALUE_COMPRESSIMAGES);
		}
	}
	else
	{
		if (! LeetRegDeleteValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_COMPRESSIMAGES) )
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Failed to delete HKCU\\%s\\%s", SN_REG_PATH_ROOT, SN_REG_VALUE_COMPRESSIMAGES);
		}
	}

	if (0 != m_nPanelCap)
	{
		DWORD dw = m_nPanelCap;

	
		if (! LeetRegSetValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_PANELCAP,
								REG_DWORD, &dw, sizeof(DWORD)) )
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Failed to save HKCU\\%s\\%s", SN_REG_PATH_ROOT, SN_REG_VALUE_PANELCAP);
		}
	}
	else
	{
		if (! LeetRegDeleteValue(HKEY_CURRENT_USER, SN_REG_PATH_ROOT, SN_REG_VALUE_PANELCAP) )
		{
			bResult = false;
			CTypedError::SafeSetSystemError(pTError, "Failed to delete HKCU\\%s\\%s", SN_REG_PATH_ROOT, SN_REG_VALUE_PANELCAP);
		}
	}

	return(bResult);
}

bool CSNConfig::LoadDevices(CTypedError *pTError, CStringList *pSLErrorDetails)
{
	CriticalSectionScoper css(&m_CS);

	bool bResult = false;

	m_pDeviceList->RemoveAndDeleteAll();

	const char *filenameIn = GetDevicesFilePath();

	if (NULL == filenameIn)
	{
		CTypedError::SafeSetMsgError(pTError, "Config returns NULL players filename");
	}
	else
	{
		if (0 != _access(filenameIn, 00))
		{
			// No config file => empty configuration.
			bResult = true;
		}
		else
		{
			g_pXMLErrorStringList->RemoveAll();

			xmlDocPtr doc = xmlParseFile(filenameIn);

			if (NULL == doc)
			{
				if (CTypedError::SafeIsNoError(pTError))
				{
					CTypedError::SafeSetMsgError(pTError, "Error reading and parsing %s", filenameIn);

					if (NULL != pSLErrorDetails)
					{
						pSLErrorDetails->AddTail(g_pXMLErrorStringList);
					}
				}

				g_pXMLErrorStringList->RemoveAll();
			}
			else
			{
				xmlNodePtr pDocRootNode = xmlDocGetRootElement(doc);

				if ((NULL == pDocRootNode) || (NULL == pDocRootNode->name))
				{
					CTypedError::SafeSetMsgError(pTError, "Empty document: %s", filenameIn);
				}
				else if (0 != strcmp((const char *)pDocRootNode->name, (const char *)SNXML_DOC_DEVICES))
				{
					CTypedError::SafeSetMsgError(pTError, "Expected root node \"%s\" but document \"%s\" has root node \"%s\".",
											(const char *)SNXML_DOC_DEVICES, filenameIn, (const char *)pDocRootNode->name);
				}
				else
				{
					bResult = true;

					// Check document version.

					xmlChar *pxcDocVersIn = xmlGetProp(pDocRootNode, SNXML_PROP_VERSION);

					int docVersIn = 0;
					int docVersDefault = atoi((const char*)SNXML_VALUE_DEVICES_VERSION);

					if (NULL != pxcDocVersIn)
					{
						docVersIn = atoi((const char *)pxcDocVersIn);
						xmlFree(pxcDocVersIn);
					}

					// Warn in the log so that if there are errors
					// the user sees it, but if not it doesn't bother them.
					if (0 == docVersIn)
					{
						CString versionWarning;

						pSLErrorDetails->AddTail("Warning: document has no version.");
						versionWarning.Format("Document = %s", filenameIn);
						pSLErrorDetails->AddTail(versionWarning);
						pSLErrorDetails->AddTail("");
					}
					else if (docVersIn > docVersDefault)
					{
						CString versionWarning;

						pSLErrorDetails->AddTail("Warning: document saved by a newer version of Super-Nudelist.");
						versionWarning.Format("Document = %s", filenameIn);
						pSLErrorDetails->AddTail(versionWarning);
						versionWarning.Format("Document version = %d ; Fully supported version = %d",
												docVersIn, docVersDefault);
						pSLErrorDetails->AddTail(versionWarning);
						pSLErrorDetails->AddTail("");
					}

					// Let's build our structures.

					bool bPrefixError = CTypedError::SafeIsNoError(pTError);

					for (xmlNodePtr pNode = pDocRootNode->children; NULL != pNode; pNode = pNode->next)
					{
						if ((!xmlIsBlankNode(pNode)) && (NULL != pNode->name))
						{
							if (0 == strcmp((const char *)SNXML_NODE_DEVICES, (const char *)pNode->name))
							{
								if (! m_pDeviceList->BuildFromXML(pTError, pSLErrorDetails, pNode) )
								{
									bResult = false;

									break;
								}
							}
						}
					}

					if (bPrefixError && CTypedError::SafeIsError(pTError))
					{
						CTypedError::SafePrefixMsgError(pTError, "Error with players file");
					}
				}

				xmlFreeDoc(doc);
				doc = NULL;
			}
		}
	}

	m_pDeviceList->ClearDirty();

	return(bResult);
}

// Set szImportFilename to NULL to load to the config file rather than an import.
bool CSNConfig::LoadLibrary(CTypedError *pTError, CStringList *pSLErrorDetails, const char *szImportFilename)
{
	CriticalSectionScoper css(&m_CS);

	bool bResult = false;

	if (NULL == szImportFilename)
	{
		m_pLibrary->RemoveAndDeleteAll();
	}

	const char *filenameIn = szImportFilename;
	
	if (NULL == filenameIn)
	{
		filenameIn = GetLibraryFilePath();
	}

	if (NULL == filenameIn)
	{
		CTypedError::SafeSetMsgError(pTError, "Config returns NULL library filenae");
	}
	else
	{
		if (0 != _access(filenameIn, 00))
		{
			// No config file => empty configuration.
			bResult = true;
		}
		else
		{
			g_pXMLErrorStringList->RemoveAll();

			xmlDocPtr doc = xmlParseFile(filenameIn);

			if (NULL == doc)
			{
				if (CTypedError::SafeIsNoError(pTError))
				{
					CTypedError::SafeSetMsgError(pTError, "Error reading and parsing %s", filenameIn);

					if (NULL != pSLErrorDetails)
					{
						pSLErrorDetails->AddTail(g_pXMLErrorStringList);
					}
				}

				g_pXMLErrorStringList->RemoveAll();
			}
			else
			{
				xmlNodePtr pDocRootNode = xmlDocGetRootElement(doc);

				if ((NULL == pDocRootNode) || (NULL == pDocRootNode->name))
				{
					CTypedError::SafeSetMsgError(pTError, "Empty document: %s", filenameIn);
				}
				else if (0 != strcmp((const char *)pDocRootNode->name, (const char *)SNXML_DOC_LIBRARY))
				{
					CTypedError::SafeSetMsgError(pTError, "Expected root node \"%s\" but document \"%s\" has root node \"%s\".",
											(const char *)SNXML_DOC_LIBRARY, filenameIn, (const char *)pDocRootNode->name);
				}
				else
				{
					bResult = true;

					// Check document version.

					xmlChar *pxcDocVersIn = xmlGetProp(pDocRootNode, SNXML_PROP_VERSION);

					int docVersIn = 0;
					int docVersDefault = atoi((const char*)SNXML_VALUE_LIBRARY_VERSION);

					if (NULL != pxcDocVersIn)
					{
						docVersIn = atoi((const char *)pxcDocVersIn);
						xmlFree(pxcDocVersIn);
					}

					// Warn in the log so that if there are errors
					// the user sees it, but if not it doesn't bother them.
					if (0 == docVersIn)
					{
						CString versionWarning;

						pSLErrorDetails->AddTail("Warning: document has no version.");
						versionWarning.Format("Document = %s", filenameIn);
						pSLErrorDetails->AddTail(versionWarning);
						pSLErrorDetails->AddTail("");
					}
					else if (docVersIn > docVersDefault)
					{
						CString versionWarning;

						pSLErrorDetails->AddTail("Warning: document saved by a newer version of Super-Nudelist.");
						versionWarning.Format("Document = %s", filenameIn);
						pSLErrorDetails->AddTail(versionWarning);
						versionWarning.Format("Document version = %d ; Fully supported version = %d",
												docVersIn, docVersDefault);
						pSLErrorDetails->AddTail(versionWarning);
						pSLErrorDetails->AddTail("");
					}

					// Let's build our structures.

					bool bPrefixError = CTypedError::SafeIsNoError(pTError);

					for (xmlNodePtr pNode = pDocRootNode->children; NULL != pNode; pNode = pNode->next)
					{
						if ((!xmlIsBlankNode(pNode)) && (NULL != pNode->name))
						{
							if (0 == strcmp((const char *)SNXML_NODE_LIBRARY, (const char *)pNode->name))
							{
								if (! m_pLibrary->BuildFromXML(pTError, pSLErrorDetails, pNode) )
								{
									bResult = false;

									break;
								}
							}
						}
					}

					if (bPrefixError && CTypedError::SafeIsError(pTError))
					{
						CTypedError::SafePrefixMsgError(pTError, "Error with library file");
					}
				}

				xmlFreeDoc(doc);
				doc = NULL;
			}
		}
	}

	set<CSNDisc *> *pSetDiscsWithUnresolvedDevices = ResolveDevices();

	delete pSetDiscsWithUnresolvedDevices;
	pSetDiscsWithUnresolvedDevices = NULL;

	if (NULL == szImportFilename)
	{
		m_pLibrary->ClearDirty();
	}

	return(bResult);
}

bool CSNConfig::SaveDevices(CTypedError *pTError) const
{
	CriticalSectionScoper css(&m_CS);

	bool bResult = false;

	const char *filenameOut = GetDevicesFilePath();

	if (NULL == filenameOut)
	{
		CTypedError::SafeSetMsgError(pTError, "Config returns NULL players filename");
	}
	else
	{
		char *parentPath = GetParentPath(filenameOut);

		if (! RecursiveCreateDirectoryDelayedRetry(parentPath) )
		{
			CTypedError::SafeSetSystemError(pTError, "Error creating %s", parentPath);
		}
		else
		{
			xmlDocPtr doc = xmlNewDoc(SNXML_VALUE_XML_VERSION);

			if (NULL == doc)
			{
				CTypedError::SafeSetMsgError(pTError, "Failed to create XML document");
			}
			else
			{
				doc->children = xmlNewDocNode(doc, NULL, SNXML_DOC_DEVICES, NULL);
				xmlSetProp(doc->children, SNXML_PROP_VERSION, SNXML_VALUE_DEVICES_VERSION);

				m_pDeviceList->AddToXML(doc->children);

				FILE *pFileOut = fopen(filenameOut, "w");

				if (NULL == pFileOut)
				{
					CTypedError::SafeSetMsgError(pTError, "Error writing to %s", filenameOut);
				}
				else
				{
					if ((-1) == xmlDocDump(pFileOut, doc))
					{
						CTypedError::SafeSetMsgError(pTError, "Error saving to %s", filenameOut);
					}
					else
					{
						bResult = true;
					}

					fclose(pFileOut);
					pFileOut = NULL;
				}

				xmlFreeDoc(doc);
				doc = NULL;
			}
		}

		delete [] parentPath;
		parentPath = NULL;
	}

	if (bResult)
	{
		m_pDeviceList->ClearDirty();
	}

	return(bResult);
}

// Set szExportFilename to NULL to save to the config file rather than an export.
bool CSNConfig::SaveLibrary(CTypedError *pTError, const char *szExportFilename) const
{
	CriticalSectionScoper css(&m_CS);

	bool bResult = false;

	const char *filenameOut = szExportFilename;
	
	if (NULL == filenameOut)
	{
		filenameOut = GetLibraryFilePath();
	}

	if (NULL == filenameOut)
	{
		CTypedError::SafeSetMsgError(pTError, "Config returns NULL library filenae");
	}
	else
	{
		char *parentPath = GetParentPath(filenameOut);

		if (! RecursiveCreateDirectoryDelayedRetry(parentPath) )
		{
			CTypedError::SafeSetSystemError(pTError, "Error creating %s", parentPath);
		}
		else
		{
			xmlDocPtr doc = xmlNewDoc(SNXML_VALUE_XML_VERSION);

			if (NULL == doc)
			{
				CTypedError::SafeSetMsgError(pTError, "Failed to create XML document");
			}
			else
			{
				doc->children = xmlNewDocNode(doc, NULL, SNXML_DOC_LIBRARY, NULL);
				xmlSetProp(doc->children, SNXML_PROP_VERSION, SNXML_VALUE_LIBRARY_VERSION);

				m_pLibrary->AddToXML(doc->children);

				FILE *pFileOut = fopen(filenameOut, "w");

				if (NULL == pFileOut)
				{
					CTypedError::SafeSetMsgError(pTError, "Error writing to %s", filenameOut);
				}
				else
				{
					if ((-1) == xmlDocDump(pFileOut, doc))
					{
						CTypedError::SafeSetMsgError(pTError, "Error saving to %s", filenameOut);
					}
					else
					{
						bResult = true;
					}

					fclose(pFileOut);
					pFileOut = NULL;
				}

				xmlFreeDoc(doc);
				doc = NULL;
			}
		}

		delete [] parentPath;
		parentPath = NULL;
	}

	if (bResult && (NULL == szExportFilename))
	{
		m_pLibrary->ClearDirty();
	}

	return(bResult);
}

void CSNConfig::SetConfigAndTempAreaPath(const char *path)
{
	CriticalSectionScoper css(&m_CS);

	delete [] m_configAndTempPath;
	m_configAndTempPath = NULL;

	delete [] m_configPath;
	m_configPath = NULL;

	delete [] m_devicesFilePath;
	m_devicesFilePath = NULL;

	delete [] m_libraryFilePath;
	m_libraryFilePath = NULL;

	delete [] m_tempPath;
	m_tempPath = NULL;

	delete [] m_tempCSFPath;
	m_tempCSFPath = NULL;

	if ((NULL != path) && (0 < strlen(path)))
	{
		m_configAndTempPath = new char[strlen(path) + 1];
		strcpy(m_configAndTempPath, path);
	}

	if (NULL != m_configAndTempPath)
	{
		m_configPath		= JoinPaths(m_configAndTempPath,	SN_DIRECTORY_CONFIG);
		m_devicesFilePath	= JoinPaths(m_configPath,			SN_FILE_DEVICES);
		m_libraryFilePath	= JoinPaths(m_configPath,			SN_FILE_LIBRARY);
		m_tempPath			= JoinPaths(m_configAndTempPath,	SN_DIRECTORY_TEMP);
		m_tempCSFPath		= JoinPaths(m_tempPath,				SN_FILE_TEMPCSF);
	}

	if ((NULL != GetExeDirectoryPath(NULL)) && (NULL != m_configAndTempPath) \
	&&  (0 == stricmp(m_configAndTempPath, m_exeDirectoryPath)))
	{
		m_bConfigAndTempPathIsDefault = true;
	}
	else
	{
		m_bConfigAndTempPathIsDefault = false;
	}
}

bool CSNConfig::SetConfigAndTempAreaPathDefault(CTypedError *pTError)
{
	CriticalSectionScoper css(&m_CS);

	bool bResult = false;

	if (NULL != GetExeDirectoryPath(pTError))
	{
		SetConfigAndTempAreaPath(m_exeDirectoryPath);
		bResult = true;
	}

	return(bResult);
}

void CSNConfig::SetInputCCFPath(const char *path)
{
	CriticalSectionScoper css(&m_CS);

	delete [] m_inputCCFPath;
	m_inputCCFPath = NULL;

	if ((NULL != path) && (0 < strlen(path)))
	{
		m_inputCCFPath = CorrectCaseOfPath(path);
	}
}

void CSNConfig::SetOutputCCFPath(const char *path)
{
	CriticalSectionScoper css(&m_CS);

	delete [] m_outputCCFPath;
	m_outputCCFPath = NULL;

	if ((NULL != path) && (0 < strlen(path)))
	{
		m_outputCCFPath = CorrectCaseOfPath(path);
	}
}

void CSNConfig::SetCompressImages(bool bCompressImages)
{
	CriticalSectionScoper ccs(&m_CS);

	m_bCompressImages = bCompressImages;
}

void CSNConfig::SetPanelCap(int nPanelCap)
{
	CriticalSectionScoper ccs(&m_CS);

	m_nPanelCap = nPanelCap;
}

// Returns NULL on error.
const char *CSNConfig::GetExePath(CTypedError *pTError) const
{
	CriticalSectionScoper css(&m_CS);

	if (NULL == m_exePath)
	{
		TCHAR modName[_MAX_PATH + 1];

		if (0 == GetModuleFileName(NULL, modName, _MAX_PATH))
		{
			CTypedError::SafeSetSystemError(pTError, "Cannot get path to executable");
		}
		else
		{
			m_exePath = CorrectCaseOfPath(modName);
		}
	}

	return(m_exePath);
}

// Returns NULL on error.
const char *CSNConfig::GetExeDirectoryPath(CTypedError *pTError) const
{
	CriticalSectionScoper css(&m_CS);

	if (NULL == m_exeDirectoryPath)
	{
		if (NULL != GetExePath(pTError))
		{
			if (NULL == (m_exeDirectoryPath = GetParentPath(m_exePath)))
			{
				CTypedError::SafeSetMsgError(pTError, "GetParentPath(%s) failed", m_exePath);
			}
		}
	}

	return(m_exeDirectoryPath);
}

// Returns NULL on error.
const char *CSNConfig::GetExeVersionString(CTypedError *pTError) const
{
	CriticalSectionScoper css(&m_CS);

	if ((NULL == m_exeVersionString) && (NULL != GetExePath(pTError)))
	{
		DWORD verSize;
		DWORD whatTheHellIsThePointOfThisVariableWhichJustGetsSetToZero;

		// Since the documentation to GetFileVersionInfoSize() says the second argument is always set to zero,
		// they must never actually do anything useful with it because I am perfectly in my rights to use
		// and rely on GetFileVersionInfoSize() when I want to set a DWORD variable to zero.
		// Bad documentation or bad code? You decide... :)

		
		verSize = GetFileVersionInfoSize(m_exePath, &whatTheHellIsThePointOfThisVariableWhichJustGetsSetToZero);

		if (0 == verSize)
		{
			CTypedError::SafeSetSystemError(pTError, "GetFileVersionInfoSize failed");
		}
		else
		{
			BYTE *verData = new BYTE[verSize];

			if (! GetFileVersionInfo(m_exePath, NULL, verSize, verData) )
			{
				CTypedError::SafeSetSystemError(pTError, "GetFileVersionInfo failed");
			}
			else
			{
				const char *pBuffer;
				UINT bufLen;

				if(! VerQueryValue(verData, "\\VarFileInfo\\Translation", (LPVOID*)(&pBuffer), &bufLen) )
				{
					CTypedError::SafeSetMsgError(pTError, "VerQueryValue failed");
				}
				else
				{
					// Find out language of version resource
					BYTE *lang = (BYTE*)pBuffer;

					TCHAR resblk[BUFFERSIZE];
					stringFormat(resblk, sizeof(resblk), "\\StringFileInfo\\%02x%02x%02x%02x\\FileVersion", lang[1], lang[0], lang[3], lang[2]);

					if (! VerQueryValue(verData, resblk, (LPVOID*)(&pBuffer), &bufLen) )
					{
						CTypedError::SafeSetMsgError(pTError, "VerQueryValue failed");
					}
					else
					{
						int n1, n2, n3, n4;

						switch(sscanf((TCHAR *)pBuffer, "%d,%d,%d,%d", &n1, &n2, &n3, &n4))
						{
						case 1: n2 = 0; // Fall through
						case 2: n3 = 0; // Fall through
						case 3: n4 = 0; // Fall through
						case 4:
							// Scope out ubtn1-4 and resLen
							{
								CUnsignedBaseTenNumber ubtn1(n1);
								CUnsignedBaseTenNumber ubtn2(n2);
								CUnsignedBaseTenNumber ubtn3(n3);
								CUnsignedBaseTenNumber ubtn4(n4);

								size_t resLen = 4 + 1; // Four fullstops and the null.

								resLen += ubtn1.GetNumberOfDigits();
								resLen += ubtn2.GetNumberOfDigits();
								resLen += ubtn3.GetNumberOfDigits();
								resLen += ubtn4.GetNumberOfDigits();

								m_exeVersionString = new char[resLen];

								stringFormat(m_exeVersionString, resLen, "%s.%s.%s.%s",
																		ubtn1.GetAsciiRepresentation(),
																		ubtn2.GetAsciiRepresentation(),
																		ubtn3.GetAsciiRepresentation(),
																		ubtn4.GetAsciiRepresentation());

							}
							break;

						default:
							m_exeVersionString = new char[strlen(pBuffer) + 1];
							strcpy(m_exeVersionString, pBuffer);
							break;
						}
					}
				}

			}

			delete [] verData;
		}
	}

	return(m_exeVersionString);
}

// Returns NULL if none set.
const char *CSNConfig::GetConfigAndTempAreaPath(void) const
{
	CriticalSectionScoper css(&m_CS);

	char *result = m_configAndTempPath;

	return(result);
}

// Returns NULL if none set.
const char *CSNConfig::GetConfigPath(void) const
{
	CriticalSectionScoper css(&m_CS);

	char *result = m_configPath;

	return(result);
}

// Returns NULL if none set.
const char *CSNConfig::GetDevicesFilePath(void) const
{
	CriticalSectionScoper css(&m_CS);

	char *result = m_devicesFilePath;

	return(result);
}

// Returns NULL if none set.
const char *CSNConfig::GetLibraryFilePath(void) const
{
	CriticalSectionScoper ccs(&m_CS);

	char *result = m_libraryFilePath;

	return(result);
}

// Returns NULL if none set.
const char *CSNConfig::GetTempPath(void) const
{
	CriticalSectionScoper css(&m_CS);

	char *result = m_tempPath;

	return(result);
}

// Returns NULL if none set.
const char *CSNConfig::GetTempCSFPath(void) const
{
	CriticalSectionScoper css(&m_CS);

	char *result = m_tempCSFPath;

	return(result);
}

// Returns NULL if none set.
const char *CSNConfig::GetInputCCFPath(void) const
{
	CriticalSectionScoper css(&m_CS);

	char *result = m_inputCCFPath;

	return(result);
}

// Returns NULL if none set.
const char *CSNConfig::GetOutputCCFPath(void) const
{
	CriticalSectionScoper css(&m_CS);

	char *result = m_outputCCFPath;

	return(result);
}

bool CSNConfig::GetCompressImages(void) const
{
	CriticalSectionScoper ccs(&m_CS);

	bool bResult = m_bCompressImages;

	return(bResult);
}

int CSNConfig::GetPanelCap(void) const
{
	CriticalSectionScoper ccs(&m_CS);

	int nPanelCap = m_nPanelCap;

	return(nPanelCap);
}

const CSNDeviceList *CSNConfig::GetDevices(void) const
{
	return(m_pDeviceList);
}

CSNDeviceList *CSNConfig::GetDevices(void)
{
	return(m_pDeviceList);
}

const CSNLibrary *CSNConfig::GetLibrary(void) const
{
	return(m_pLibrary);
}

CSNLibrary *CSNConfig::GetLibrary(void)
{
	return(m_pLibrary);
}
