// SNHardButtonDefinition.cpp: implementation of the CSNHardButtonDefinition class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SNConfig.h"
#include "SNHardButtonDefinition.h"
#include "CSFSection.h"
#include "CSFPath.h"
#include "CSFAbstractTree.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

void CSNHardButtonDefinition::init(void)
{
	m_pLabel = NULL;
	m_pJumpPanel = NULL;
	m_actionButtonPaths.SetSort(false);
	m_actionButtonPaths.SetRequireUniqueNames(false);
}

CSNHardButtonDefinition::CSNHardButtonDefinition()
{
	init();
}

CSNHardButtonDefinition::CSNHardButtonDefinition(const char *szNamePronto, bool bHasLabel)
{
	init();

	m_name.SetPronto(szNamePronto);

	if (bHasLabel)
	{
		m_pLabel = new CProntoString;
	}
}

CSNHardButtonDefinition::~CSNHardButtonDefinition()
{
	delete m_pLabel;
	SetJumpPanel(NULL);
}

const char *CSNHardButtonDefinition::getTypeName(void) const
{
	return("hard button definition");
}

bool CSNHardButtonDefinition::IsValid(CTypedError *pTError) const
{
	if (!m_name.IsANonEmptyString())
	{
		CTypedError::SafeSetMsgError(pTError, "%s has no name", getTypeName());
		return(false);
	}

	return(true);
}

// Returns < 0 if this <  pHBD
// Returns   0 if this == pHBD
// Returns > 0 if this >  pHBD
int CSNHardButtonDefinition::CompareForSort(const CSNHardButtonDefinition *pHBD) const
{
	return(m_name.Compare(pHBD->GetName(), false));
}

void CSNHardButtonDefinition::Set(const CSNHardButtonDefinition *pHBD)
{
	if (this != pHBD)
	{
		m_name.SetPronto(pHBD->m_name.GetPronto(false));

		if (NULL == pHBD->m_pLabel)
		{
			delete m_pLabel;
			m_pLabel = NULL;
		}
		else
		{
			if (NULL == m_pLabel)
			{
				m_pLabel = new CProntoString();
			}

			m_pLabel->SetPronto(pHBD->m_pLabel->GetPronto(false));
		}

		this->SetJumpPanel(pHBD->m_pJumpPanel);

		m_actionButtonPaths.Set(pHBD->m_actionButtonPaths);
	}
}

bool CSNHardButtonDefinition::IsNonEmpty(void) const
{
	if (NULL != m_pJumpPanel)
	{
		return(true);
	}

	if (NULL != m_actionButtonPaths.GetFirst())
	{
		return(true);
	}

	return(false);
}

bool CSNHardButtonDefinition::ResolveCSFPaths(CTypedError *pTError, const CCSFAbstractTree *pAT, const char *ownerType, const char *ownerName)
{
	bool bResult = true;

	bool bPrefixErrors = CTypedError::SafeIsNoError(pTError);

	if (NULL != m_pJumpPanel)
	{
		if (! m_pJumpPanel->ResolveCSFPath(pTError, true, pAT, getTypeName(),
										GetName()->GetPrintable(true), "jump panel") )
		{
			bResult = false;
		}
	}

	if (bResult)
	{
		for (CCSFButtonPath *pBut = m_actionButtonPaths.GetNonConst(m_actionButtonPaths.GetFirst()); NULL != pBut; pBut = m_actionButtonPaths.GetNonConst(m_actionButtonPaths.GetNext(pBut)))
		{
			if (! pBut->ResolveCSFPath(pTError, true, pAT, getTypeName(),
											GetName()->GetPrintable(true), "aliased button") )
			{
				bResult = false;
				break;
			}
		}
	}

	if ((!bResult) && bPrefixErrors)
	{
		CTypedError::SafePrefixMsgError(pTError, "In %s \"%s\"", ownerType, ownerName);
	}

	return(bResult);
}

const CProntoString *CSNHardButtonDefinition::GetName(void) const
{
	return(&m_name);
}

// GetLabel returns NULL if the hard button has no label (ever).
CProntoString *CSNHardButtonDefinition::GetLabel(void)
{
	return(m_pLabel);
}

const CProntoString *CSNHardButtonDefinition::GetLabel(void) const
{
	return(m_pLabel);
}

const CCSFPanelPath *CSNHardButtonDefinition::GetJumpPanel(void) const
{
	return(m_pJumpPanel);
}

void CSNHardButtonDefinition::SetJumpPanel(const CCSFPanelPath *pJumpPanel)
{
	delete m_pJumpPanel;
	m_pJumpPanel = NULL;

	if (NULL != pJumpPanel)
	{
		m_pJumpPanel = new CCSFPanelPath();
		m_pJumpPanel->Set(pJumpPanel);
	}
}

CCSFButtonPathList *CSNHardButtonDefinition::GetActionButtonPaths(void)
{
	return(&m_actionButtonPaths);
}

const CCSFButtonPathList *CSNHardButtonDefinition::GetActionButtonPaths(void) const
{
	return(&m_actionButtonPaths);
}

xmlNodePtr CSNHardButtonDefinition::AddToXML(xmlNodePtr parentNode) const
{
	xmlNodePtr hbdNode = xmlNewChild(parentNode, NULL, SNXML_NODE_HARDBUTTONDEFINITION, NULL);
	xmlSetProp(hbdNode, SNXML_PROP_NAME, (const xmlChar *)GetName()->GetEscaped(true));

	if (NULL != GetLabel())
	{
		xmlSetProp(hbdNode, SNXML_PROP_LABEL, (const xmlChar *)GetLabel()->GetEscaped(true));
	}

	if (NULL != m_pJumpPanel)
	{
		xmlNodePtr jpNode = m_pJumpPanel->AddToXML(hbdNode);
		xmlSetProp(jpNode, SNXML_PROP_NAME, SNXML_VALUE_JUMPPANEL);
	}

	for (const CCSFButtonPath *pBut = m_actionButtonPaths.GetFirst(); NULL != pBut; pBut = m_actionButtonPaths.GetNext(pBut))
	{
		xmlNodePtr lbNode = pBut->AddToXML(hbdNode);
		xmlSetProp(lbNode, SNXML_PROP_NAME, SNXML_VALUE_BUTTONALIAS);
	}

	return(hbdNode);
}

bool CSNHardButtonDefinition::BuildFromXML(CTypedError *pTError, CStringList *pSLErrorDetails, xmlNodePtr pNode, bool bIgnoreLabelIfNoneAlready)
{
	bool bResult = true;

	xmlChar *pXC;
	xmlChar *pXCType;
	xmlChar *pXCName;

	if (NULL != (pXC = xmlGetProp(pNode, SNXML_PROP_NAME)))
	{
		m_name.SetEscaped((const char *)pXC);
		xmlFree(pXC);
	}

	if (NULL != (pXC = xmlGetProp(pNode, SNXML_PROP_LABEL)))
	{
		if ((NULL == m_pLabel) && (!bIgnoreLabelIfNoneAlready))
		{
			m_pLabel = new CProntoString();
		}

		if (NULL != m_pLabel)
		{
			m_pLabel->SetEscaped((const char *)pXC);
		}

		xmlFree(pXC);
	}

	for (xmlNodePtr pNodeChild = pNode->children; NULL != pNodeChild; pNodeChild = pNodeChild->next)
	{
		if ((!xmlIsBlankNode(pNodeChild)) && (NULL != pNodeChild->name))
		{
			if (0 == strcmp((const char *)SNXML_NODE_CSFPATH, (const char *)pNodeChild->name))
			{
				if (NULL != (pXCType = xmlGetProp(pNodeChild, SNXML_PROP_TYPE)))
				{
					if (NULL != (pXCName = xmlGetProp(pNodeChild, SNXML_PROP_NAME)))
					{
						if (0 == strcmp((const char *)SNXML_VALUE_PANEL, (const char *)pXCType))
						{
							CCSFPanelPath *pPanelPath = new CCSFPanelPath();

							if (! pPanelPath->BuildFromXML(pTError, pSLErrorDetails, pNodeChild) )
							{
								bResult = false;
							}
							else if (0 == strcmp((const char *)SNXML_VALUE_JUMPPANEL, (const char *)pXCName))
							{
								SetJumpPanel(pPanelPath);
							}

							delete pPanelPath;
						}
						else if (0 == strcmp((const char *)SNXML_VALUE_BUTTON, (const char *)pXCType))
						{
							CCSFButtonPath *pButtonPath = new CCSFButtonPath();

							if (! pButtonPath->BuildFromXML(pTError, pSLErrorDetails, pNodeChild) )
							{
								bResult = false;
							}
							else if (0 == strcmp((const char *)SNXML_VALUE_BUTTONALIAS, (const char *)pXCName))
							{
								if (NULL != m_actionButtonPaths.AddBelowAndOwn(pTError, pButtonPath, NULL))
								{
									pButtonPath = NULL; // Owned.
								}
								else
								{
									bResult = false;
								}
							}

							delete pButtonPath;
						}

						xmlFree(pXCName);
					}

					xmlFree(pXCType);
				}
			}
		}
	}

	return(bResult);
}

void CSNHardButtonDefinition::AddToDeviceCSFSection(CCSFSection *pDeviceCSFSection) const
{
	CString secTitle;
	secTitle = m_name.GetPronto(true);
	secTitle += "(";

	if (NULL != m_pLabel)
	{
		secTitle += "\"";
		secTitle += m_pLabel->GetEscaped(true);
		secTitle += "\"";
	}

	secTitle += ")";

	CCSFSection *pCSFSection = pDeviceCSFSection->AddSubSection();
	pCSFSection->SetTitleEscaped(secTitle);

	for (const CCSFButtonPath *pButPath = m_actionButtonPaths.GetFirst(); NULL != pButPath; pButPath = m_actionButtonPaths.GetNext(pButPath))
	{
		pCSFSection->AddEscapedTitleOnlySubSection((*(pButPath->GetATButtons()->begin()))->GetCSFAliasString());
	}

	if (NULL != m_pJumpPanel)
	{
		pCSFSection->AddEscapedTitleOnlySubSection((*(m_pJumpPanel->GetATPanels()->begin()))->GetCSFJumpString());
	}
}
