// SNList.h: interface for the CSNList class.
//
//////////////////////////////////////////////////////////////////////

// version 1.0 - 19/Oct/2000 - Leo Davidson.
// * Initial version.

// Notes:
// * Currently not suitable for simultaneous access from multiple threads.
// * Insert and remove operations are linear for now.
//   (Seems to scale reasonably well to 1000 items, but with 10000 it is
//   far too slow. I'm not expecting lists anywhere near that size in
//	 current projects.)

#if !defined(AFX_SNLIST_H__2E8770F9_C7A3_4A46_B21E_D5E142E9DBD0__INCLUDED_)
#define AFX_SNLIST_H__2E8770F9_C7A3_4A46_B21E_D5E142E9DBD0__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


/////////////////////////////////////////////////////////////////////////

class CSNListObject
{
public:
	CSNListObject() : m_indexInOwnerList(0), m_bUniqueName(true) {}
	virtual ~CSNListObject() {}

	bool IsUniqueName(void) const { return(m_bUniqueName); }

protected:
	unsigned int m_indexInOwnerList;
	bool m_bUniqueName;
};

/////////////////////////////////////////////////////////////////////////
//
// TListObject should be a class derived from CSNListObject.
// TListObject should declare CSNList<TListObject> its friend.
// TListObject should define:
//
//	bool IsValid(CTypedError *pTError) const;
//
//	// Returns < 0 if this <  pObject
//	// Returns   0 if this == pObject
//	// Returns > 0 if this >  pObject
//	int CompareForSort(const TListObject *pObject) const;
//
//  // Used for updating existing elements only.
//  void Set(const TListObject *pObject);
//
//	// Type name should be lowercase, singular.
//	const char *getTypeName(void) const;
//
/////////////////////////////////////////////////////////////////////////

template<class TListObject> class CSNUpdatableList;
template<class TListObject> class CSNSetableList;

template<class TListObject> class CSNList
{
	friend CSNUpdatableList<TListObject>;
	friend CSNSetableList<TListObject>;

public:
	CSNList();
	virtual ~CSNList();

	// To improve performance, reserve space before large insertions.
	// If bAdditional is true, readies the list for allocSize new items to
	// be added, else makes sure the list is ready for at least allocSize
	// elements in total.
	void ReserveAllocation(unsigned int allocSize, bool bAdditional);

	void FreeExcessAllocation(void);

	enum GrowType
	{
		GROW_ADD,
		GROW_MULTIPLY
	};

	// Allows you to change the way the list grows and shrinks automatially.
	// If you set GROW_MULTIPLY you may want to use specify a non-zero
	// recommendedMinSize to avoid lots of reallocations when adding and removing
	// small lists. You can always call FreeExcessAllocation when the list contents are
	// more stable.
	void SetAllocationRule(GrowType growType,
						   unsigned int growSizeMultipliedByTen,
						   unsigned int recommendedMinSize);

	// Sorting is disabled by default.
	void SetSort(bool bSort);
	bool GetSort(void) const;

	// Unique names are required by default.
	// Any duplicate names in the list remain when you turn unique names back on.
	void SetRequireUniqueNames(bool bReqUniqueNames);
	bool GetRequireUniqueNames(void) const;

	virtual void ClearDirty(void);
	virtual bool IsDirty(void) const;

	void RemoveAndDeleteAll(void);

	unsigned int GetNumElements(void) const;

	// Removing may cause (some) excess allocated space to be freed.
	void RemoveAndDelete(const TListObject *pToRemove);

	// Returns NULL on failure.
	TListObject *RemoveAndReleaseOwnership(const TListObject *pToRemove);

	// Move functions do nothing when the list is sorted.
	void MoveUp(const TListObject *pToMove);
	void MoveDown(const TListObject *pToMove);

	// Keeps the list sorted if it is, in which case pToAddBelow is ignored.
	// Give NULL pToAddBelow to add to the bottom of the list.
	// Returns NULL on error, else a pointer to the added object.
	// Iff fails the caller must delete pInObject themselves.
	const TListObject *AddBelowAndOwn(CTypedError *pTError, TListObject *pInObject, const TListObject *pToAddBelow);

	// Returns NULL if not found.
	// If pFindAfter is NULL the search starts with the first element.
	const TListObject *GetByNamePronto(const char *szNamePronto, const TListObject *pFindAfter) const;

	// Iteration functions return NULL when there is no such element.
	const TListObject *GetFirst(void) const;
	const TListObject *GetLast(void) const;
	const TListObject *GetNext(const TListObject *pCurrent) const;
	const TListObject *GetPrevious(const TListObject *pCurrent) const;

	// Returns NULL on failure.
	TListObject *GetNonConst(const TListObject *pConst);

protected:
	// pToExclude may be NULL.
	// Set startIdx to zero to search entire list.
	// If is safe to call with startIdx indicating past the end of the list.
	TListObject *getByNameExcluding(const CProntoString *pName,
										const TListObject *pToExclude,
										unsigned int startIdx) const;

private:
	// Will not shrink the array too small to hold all current elements.
	bool resizeArray(unsigned int newSize);

	// Must always increase / decrease the size by at least one.
	// Does not actually change the size, just calculates a new one.
	// When shrinking, it does not check against the element count.
	unsigned int calcSize(unsigned int oldSize, bool bGrow) const;

	static int qSortCompare(const void *arg1, const void *arg2);

	// Doesn't make sense to call when the list is not sorted.
	// Will return index one past the end of the buffer if no item is bigger.
	unsigned int getIndexFirstGreater(const TListObject *pToOwn);

	void removeElement(unsigned int idx, bool bDeleteElement, bool bAllowArrayShrink);

	bool insertElement(CTypedError *pTError, TListObject *pToOwn, unsigned int idx);

protected:
	bool m_bDirty;
private:
	bool m_bSort;
	bool m_bRequireUniqueNames;

	GrowType m_growType;
	unsigned int m_growSizeMultipliedByTen;
	unsigned int m_recommendedMinSize;

	unsigned int m_allocatedSize;
	unsigned int m_numElements;

	TListObject **m_array;
};

template<class TListObject> class CSNUpdatableList : public CSNList<TListObject>
{
public:
	CSNUpdatableList() {}
	virtual ~CSNUpdatableList() {}

	// Updating is guarnateed not to change the object's address.
	// Resorts the list if needed, else the element's position is unchanged.
	// Returns NULL on error, else a pointer to the updated object.
	// (Will *NOT* fail or detect if pInObject is valid but pToUpdate->Set(pInObject)
	// makes pToUpdate invalid.)
	const TListObject *Update(CTypedError *pTError, const TListObject *pInObject, const TListObject *pToUpdate);

	// UpdateAndMove does what you could do with RemoveAndTakeOwnership an update
	// if the item and then AddBelowAndOwn, except as one atomic operation which
	// will either succeed or do nothing, exceptional situations (low-memory, etc.) aside.
	// (Will *NOT* fail or detect if pInObject is valid but pToUpdateInSource->Set(pInObject)
	// makes pToUpdateInSource invalid.)
	// Safe to call with pTargetList the same as this list.
	// Call with InObject NULL to move without updating the element.
	// Call with pTargetList NULL to move within the same list.
	// Call with pToAddBelowInTarget NULL to add to the bottom of the target list.
	// If the target list is sorted, pToAddBelowInTarget is ignored.
	const TListObject *UpdateAndMove(CTypedError *pTError, CSNList<TListObject> *pTargetList, const TListObject *pInObject, const TListObject *pToUpdateInSource, const TListObject *pToAddBelowInTarget);
};

template<class TListObject> class CSNSetableList : public CSNUpdatableList<TListObject>
{
public:
	CSNSetableList() {}
	virtual ~CSNSetableList() {}

	// Empties this list and copies all elements from the other list to this one.
	// Also copies the sort and unique-names flags and memory allocation parameters.
	// Sets dirty flag.
	void Set(const CSNSetableList<TListObject>&listToSet);
};

/////////////////////////////////////////////////////////////////////////

#endif // !defined(AFX_SNLIST_H__2E8770F9_C7A3_4A46_B21E_D5E142E9DBD0__INCLUDED_)
