#ifndef __TYPEDERROR_H_
#define __TYPEDERROR_H_

#define TYPEDERROR_WINSOCK	FALSE
#define	TYPEDERROR_WNET		FALSE

#define TYPEDERROR_BUFFERSIZE 5120

#include <lmerr.h>

#if TYPEDERROR_WINSOCK
//#include <winsock2.h>
#endif

#define TE_NOERRORRECORDED "<no error recored>"
#define TE_ERRORSTRINGPARTSEPARATOR ": "


// This class is not designed for simultaneous access from multiple threads.


// Functions which set the error will be ignored if an error is already set.
// Call Reset when you want to be able to set another error.
// This behaviour allows you to call functions which may set error codes after something earlier has failed without
// going over the original error message.

// v1.1 20-Dec-1999 Leo Davidson
// * Added GetFullErrorText() which calculates the size of and returns a new buffer for you, as all good APIs should.
// * Renamed to CTypedError (was TypedError).
// * Added the SafeSet static functions, removing the old Set functions.
// * Moved the ErrorType typedef inside the class.
// * Added support for (extended) WNet errors.
// * SafeCopyTError copies another TError to this one while the new SafeSetTError appends the message from
//   another TError to your message, similar to how the other functions append a system message to yours.
// v1.2 26-Jan-2000 Leo Davidson
// * Added C-runtime error reporting. (Untested)
// v1.3 31-Jan-2000 Leo Davidson
// * Replaced GetFullErrorText functions with SafeGetFullErrorText versions.
// v1.4 03-August-2000 Leo Davidson
// * Added support for reporting RPC status errors, although they seem to be a subset of the Windows errors (FFS!).
// * RPC and Winsock errors can be turned off using defines at the top.
// * Removed GetFullErrorText functions which didn't allocate the buffer for the caller.
// * Changed error type from int to DWORD.
// * Added SafeReset.
// * Added SafeGetFullErrorTextTempBuffer and GetFullErrorTextTempBuffer.
// v1.7 15-August-2000 Leo Davidson
// * Removed explicit RPC error support since the RPC errors are a subset of the Windows system errors. (doh!)
// * Slightly changed how Winsock support is enabled/disabled.
// * Moved code to new TypedError.cpp for a more readable header.
// * All SafeSet* functions can take a NULL or empty msg value now.
// * Hopefully improved getFullErrorText's handling of empty messages and code clarity.
// * GetSystemErrorText and GetWinSockErrorText are now static members. (Used to be in Gloabl.cpp)
// v1.8 25-August-2000 Leo Davidson
// * Fixed compilation of WinSock-related code.
// * SafeGetFullErrorTextTempBuffer now takes a pointer to a const CTypedError and casts away the const.
// v1.9 27-August-2000 Leo Davidson
// * Added SafeIsNoError and SafeIsError which default to no error. Old unsafe functions made protected.
// * Added important note above SafePrefixMsgError.

class CTypedError
{
public:
	enum ErrorType
	{
		ERROR_NONE,
		ERROR_SYSTEM,
		ERROR_WINSOCK,
		ERROR_MESSAGEONLY,
		ERROR_CRUNTIME,
	} ;

public:
	CTypedError(void);
	virtual ~CTypedError(void);

	DWORD GetError(void) const;
	ErrorType GetErrorType(void) const;
	const char *GetErrorMessage(void) const; // This is not the FULL error message. See other functions...
	const char *GetPrependMessage(void) const;

	// Set a message which is prepended to whatever message is generated by the TError.
	// Each call overwrites the previous prepend message.
	// If you just want to prefix something to the message that is currently set, see SafePrefixMsgError.
	void SetPrependMessage(const char *preMsg);

	// Static safe-set functions. These check the CTypedError pointer for NULL so that you can easily
	// and safely provide support for an optional CTypedError without checking it all the time.

	// Reset the TError to containing no error.
	static void SafeReset(CTypedError *pTError);

	// Defaults to true if pTError is NULL.
	static bool SafeIsNoError(CTypedError *pTError);

	// Defaults to false if pTError is NULL.
	static bool SafeIsError(CTypedError *pTError);

	// Copies one TError to another, if the destination does not already contain an error.
	// See also SafeSetTError which is similar.
	static void SafeCopyTError(CTypedError *pTErrorDestination, const CTypedError *pTErrorSource,
								bool bCopyPrependMessage);

	// Returns full error text in a temporary buffer which is deleted when the typed error is deleted or changed.
	static const char *SafeGetFullErrorTextTempBuffer(const CTypedError *pTError);

	// Returns a new buffer. delete[] when finished with it.
	static char *SafeGetFullErrorTextCopy(const CTypedError *pTError);

	// Prefixes your message to the existing message in the TError.
	// This is different to calling SetPrependMessage() which sets a *single* prefix buffer which
	// appears in front of any messages generated by the object.
	// Calling SafePrefixMsgError() multiple times will add each of your messages to the front
	// of whatever is there while calling SetPrependMessage() multiple times will result in
	// only the last call having any effect.
	// SetPrependMessage() should generally only be called by the owner of the TError while
	// SafePrefixMsgError() is mainly intended for functions which are generating errors.
	// If you call SafePrefixMsgError() and the TError has no error set it is the same as
	// calling SafeSetMsgError().
	// Important Note: Make sure you only call SafePrefixMsgError when you know that there was
	// *************** no error stored beforehand, i.e. that your function set the error.
	// *************** (If a CTypedError already has an error set, setting another one does
	// *************** nothing until after a reset has been done.)
	static void SafePrefixMsgError(CTypedError *pTError, const char *msg, ...);

	static void SafeSetMsgError(CTypedError *pTError, const char *msg, ...);

	static void SafeSetSystemError(CTypedError *pTError, const char *msg, ...);
	static void SafeSetSystemErrorNum(CTypedError *pTError, int error, const char *msg, ...);

	static void SafeSetCRuntimeError(CTypedError *pTError, const char *msg, ...);
	static void SafeSetCRuntimeErrorNum(CTypedError *pTError, int error, const char *msg, ...);

#if TYPEDERROR_WINSOCK
	static void SafeSetWSError(CTypedError *pTError, const char *msg, ...);
	static void SafeSetWSErrorNum(CTypedError *pTError, int error, const char *msg, ...);
#endif // TYPEDERROR_NOWINSOCK

	// There is no SafeSetWNetErrorNum().
	// If the error isn't ERROR_EXTENDED_ERROR it's a normal system error.
	// If the error is ERROR_EXTENDED_ERROR you can only get the extended error information
	// with a call to WNetGetLastError (via SafeSetWNetError()) or the information may be lost.

	static void SafeSetWNetError(CTypedError *pTError, const char *msg, ...);

	// Appends the message from another TError to your message, similar to how the other functions
	// append a system message to yours.
	// See also SafeCopyTError which is similar.
	static void SafeSetTError(CTypedError *pTError, const CTypedError *pTErrorToSet, const char *msg, ...);

	// Basic error number to error message functions.

	static LPTSTR GetSystemErrorText(DWORD err, LPTSTR lpszBuf, DWORD dwSize);

#if TYPEDERROR_WINSOCK
	static char *GetWinSockErrorText(int errNum, char *buffer, int bufferSize);
#endif // TYPEDERROR_WINSOCK

private:
	DWORD		m_error;			// Last error number.
	ErrorType	m_errorType;		// Context of the last error.
	char	   *m_errorMessage;		// Message about the error.
	char	   *m_prependMessage;	// Text to be prepended to error messages.

	char	   *m_tempFullErrorText;	// Used to return error message to some functions.

	void sem(const char *nMsg);
	void spm(const char *nMsg);

protected:
	void reset(void);
	bool isNoError(void) const;
	bool isError(void) const;

	void set(const char *msg, DWORD err, ErrorType errorType, va_list args);

	// Returns full error text in a temporary buffer which is deleted when the typed error is deleted or changed.
	const char *getFullErrorTextTempBuffer(void);

	// Returns a new buffer. delete[] when finished with it.
	char *getFullErrorText(void) const;
};

#endif
